# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Processing workflow for config-based processing."""

import logging
from copy import deepcopy
from glob import glob
from os.path import exists
from os.path import basename
from os import getpid
from datetime import datetime
from pyaml_env import parse_config

from geoips.commandline.args import check_command_line_args
from geoips.filenames.base_paths import PATHS as gpaths
from geoips.utils.memusg import PidLog

from geoips.geoips_utils import output_process_times
from geoips.dev.product import (
    get_covg_from_product,
    get_covg_args_from_product,
    get_required_variables,
)
from geoips.xarray_utils.data import (
    sector_xarrays,
    combine_preproc_xarrays_with_alg_xarray,
)
from geoips.filenames.duplicate_files import remove_duplicates
from geoips.geoips_utils import replace_geoips_paths
from geoips.utils.context_managers import import_optional_dependencies

with import_optional_dependencies(loglevel="info"):
    """Attempt to import a package and print to LOG.info if the import fails."""
    from geoips_db.utils.database_writes import (
        write_to_database,
        flag_product_as_deleted,
        write_stats_to_database,
    )

# Old interfaces (YAML, will migrate to new soon)
from geoips.dev.output_config import (
    get_output_formatter_kwargs,
    get_output_formatter,
    get_minimum_coverage,
    produce_current_time,
)

# New interfaces
from geoips.interfaces import interpolators
from geoips.interfaces import output_formatters
from geoips.interfaces import readers
from geoips.interfaces import products
from geoips.interfaces import sector_adjusters

# Collect functions from single_source (should consolidate these somewhere)
from geoips.plugins.modules.procflows.single_source import (
    process_sectored_data_output,
    process_xarray_dict_to_output_format,
    pad_area_definition,
    add_filename_extra_field,
    get_area_defs_from_command_line_args,
    plot_data,
    combine_filename_extra_fields,
    get_alg_xarray,
    verify_area_def,
    remove_unsupported_kwargs,
    get_unique_dataset_key,
)

# Moved to top-level errors module, fixing issue #67
from geoips.errors import CoverageError, PluginError

PMW_NUM_PIXELS_X = 1400
PMW_NUM_PIXELS_Y = 1400
PMW_PIXEL_SIZE_X = 1000
PMW_PIXEL_SIZE_Y = 1000

LOG = logging.getLogger(__name__)

interface = "procflows"
family = "standard"
name = "config_based"


# get geoips version
try:
    geoips_version = gpaths["GEOIPS_VERS"]
except KeyError:
    LOG.warning("No geoips system defined, setting geoips version to 0.0.0")
    geoips_version = "0.0.0"


def update_output_dict_from_command_line_args(output_dict, command_line_args=None):
    """Update output dict from command line args."""
    if command_line_args is None:
        LOG.info(
            "SKIPPING command_line_args not specified, returning original output_dict"
        )
        return output_dict

    final_output_dict = output_dict.copy()
    for cmdline_fld_name in [
        "filename_formatter_kwargs",
        "metadata_filename_formatter_kwargs",
    ]:
        # Skip fields that are NOT in command_line_args
        if (
            cmdline_fld_name not in command_line_args
            or command_line_args[cmdline_fld_name] is None
        ):
            LOG.info("SKIPPING %s argument not specified command line")
            continue

        # Convert filename_formatter_kwargs and metadata_filename_formatter_kwargs to
        # their plural counterparts
        # Ensure we copy the command_line_args so we do not inadvertently modify
        # the original command line arguments. This should likely be refactored.
        if cmdline_fld_name == "filename_formatter_kwargs":
            output_fld_name = "filename_formatters_kwargs"
            output_fld_val = {"all": deepcopy(command_line_args[cmdline_fld_name])}
        elif cmdline_fld_name == "metadata_filename_formatter_kwargs":
            output_fld_name = "metadata_filename_formatters_kwargs"
            output_fld_val = {"all": deepcopy(command_line_args[cmdline_fld_name])}
        else:
            output_fld_name = cmdline_fld_name
            output_fld_val = deepcopy(command_line_args[cmdline_fld_name])

        # If the current command line field is not in the output dict at all, just
        # add the whole thing.
        if output_fld_name not in output_dict:
            LOG.info(
                'ADDING output_dict "%s" from command_line_args: %s',
                cmdline_fld_name,
                output_fld_val,
            )
            final_output_dict[output_fld_name] = output_fld_val

        # If the special key 'all' is in the desired output_dict field,
        # and it is NOT currently in the
        # output_dict, then add the entire "all" field
        elif "all" in output_fld_val and "all" not in output_dict[output_fld_name]:
            LOG.warning(
                'REPLACING output_dict "%s" *all*with command_line_args: %s',
                output_fld_name,
                output_fld_val,
            )
            final_output_dict[output_fld_name]["all"] = output_fld_val["all"]

        # If the special key "all" is in the desired output_dict field,
        # but IS currently in the output_dict, then add the individual fields
        # within the command line dictionary
        elif "all" in output_fld_val:
            LOG.warning(
                'REPLACING output_dict "%s" *fields* with command_line_args: %s',
                output_fld_name,
                output_fld_val,
            )
            for kwarg_name in output_fld_val["all"]:
                final_output_dict[output_fld_name]["all"][kwarg_name] = output_fld_val[
                    "all"
                ][kwarg_name]

        # Otherwise this is a normal argument, just replace it.
        else:
            LOG.warning(
                'REPLACING entire output_dict "%s" with command_line_args: %s',
                output_fld_name,
                output_fld_val,
            )
            final_output_dict[cmdline_fld_name] = output_fld_val

    return final_output_dict


def get_required_outputs(config_dict, sector_type):
    """Get only the required outputs from the current sector_type."""
    return_dict = {}
    for output_type, output_dict in config_dict["outputs"].items():
        # If the current output type does not require the current sector_type, skip
        if output_dict["requested_sector_type"] != sector_type:
            continue
        return_dict[output_type] = output_dict

    return return_dict


def get_bg_xarray(
    sect_xarrays,
    area_def,
    prod_plugin,
    resampled_read=False,
    window_start_time=None,
    window_end_time=None,
):
    """Get background xarray.

    Parameters
    ----------
    sect_xarrays: dict of xarray.Dataset
        dictionary of xarray Datasets to pull appropriate background xarray from.
        This may include multiple products / variables.
    area_def: pyresample.AreaDefinition
        Spatial region required in the final xarray Datasets.
    prod_plugin: ProductPlugin
        GeoIPS Product Plugin obtained through interfaces.products.get_plugin("name").
    resampled_read: bool, default=False
        Specify whether a resampled read is required, needed for datatypes that
        will be read within "get_alg_xarray"
    window_start_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!
    window_end_time: datetime.datetime, default=None
        If specified, sector temporally between window_start_time and window_end_time.
        hours_before_sector_time and hours_after_sector_time are ignored if
        window start/end time are set!

    Returns
    -------
    alg_xarray: xarray.Dataset
        xarray Dataset containing the data needed to produce the background for
        overlay imagery.
    """
    interp_plugin = None
    LOG.interactive(
        "Getting background xarray dataset for product '%s'", prod_plugin.name
    )
    if "interpolator" in prod_plugin["spec"]:
        interp_plugin = interpolators.get_plugin(
            prod_plugin["spec"]["interpolator"]["plugin"]["name"]
        )
        interp_args = prod_plugin["spec"]["interpolator"]["plugin"]["arguments"]
        interp_args = remove_unsupported_kwargs(interp_plugin, interp_args)

    alg_xarray = None

    # If this is a preprocessed data file with the final product in it,
    # just pull the final product. Must take out METADATA dataset!
    if (
        len(set(sect_xarrays.keys()).difference({"METADATA"})) == 1
        and prod_plugin.name in list(sect_xarrays.values())[0].variables
    ):
        sect_xarray = list(sect_xarrays.values())[0]

        # Efficiency hit with to_masked_array
        # LOG.info('Min/max %s %s / %s',
        #          product_name,
        #          sect_xarray[product_name].to_masked_array().min(),
        #          sect_xarray[product_name].to_masked_array().max())

        LOG.interactive(
            "Interpolating %s xarray with args %s",
            sect_xarray.source_name,
            str(interp_args),
        )
        alg_xarray = interp_plugin(
            area_def, sect_xarray, alg_xarray, varlist=[prod_plugin.name], **interp_args
        )

        # Efficiency hit with to_masked_array
        # LOG.info('Min/max interp %s %s / %s',
        #          product_name,
        #          alg_xarray[product_name].min(),
        #          alg_xarray[product_name].max())

    # If this is a raw datafile, pull the required variables for applying the given
    # algorithm, and generate the product array.
    else:
        pad_area_def = pad_area_definition(area_def)
        # Ensure pre-processed and raw look the same - this requires applying algorithm
        # to padded sectored data, since that is what is written out to the
        # pre-processed netcdf file, then interpolating to the desired area definition.

        sect_xarray = get_alg_xarray(
            sect_xarrays,
            pad_area_def,
            prod_plugin,
            resampled_read=resampled_read,
            window_start_time=window_start_time,
            window_end_time=window_end_time,
        )
        LOG.interactive(
            "Interpolating %s xarray with varlist %s",
            sect_xarray.source_name,
            prod_plugin.name,
        )
        alg_xarray = interp_plugin(
            area_def, sect_xarray, alg_xarray, varlist=[prod_plugin.name]
        )

    alg_xarray.attrs["registered_dataset"] = True
    alg_xarray.attrs["area_definition"] = area_def
    if prod_plugin.name in alg_xarray.variables:
        alg_xarray = add_filename_extra_field(
            alg_xarray, "background_data", f"bg{prod_plugin.name}"
        )

    return alg_xarray


def get_resampled_read(
    config_dict,
    area_defs,
    area_def_id,
    sector_type,
    reader_plugin,
    reader_kwargs,
    fnames,
    variables,
):
    """Return dictionary of xarray datasets for a given area def.

    Xarrays resampled to area_def
    """
    return get_sectored_read(
        config_dict,
        area_defs,
        area_def_id,
        sector_type,
        reader_plugin,
        reader_kwargs,
        fnames,
        variables,
    )


def get_sectored_read(
    config_dict,
    area_defs,
    area_def_id,
    sector_type,
    reader_plugin,
    reader_kwargs,
    fnames,
    variables,
):
    """Return dictionary of xarray datasets for a given area def.

    Xarrays sectored to area_def
    """
    area_def = area_defs[area_def_id][sector_type]["area_def"]

    if "primary_sector" in config_dict:
        primary_sector_type = area_defs[area_def_id][config_dict["primary_sector"]]
        pad_area_def = primary_sector_type["area_def"]
    else:
        pad_area_def = pad_area_definition(area_def)
    try:
        xobjs = reader_plugin(
            fnames,
            metadata_only=False,
            chans=variables,
            area_def=pad_area_def,
            **reader_kwargs,
        )
    # geostationary satellites fail with IndexError when the area_def does not intersect
    # the data.  Just skip those.  We need a better method for handling this generally,
    # but for now skip IndexErrors.
    except IndexError as resp:
        LOG.error("%s SKIPPING no coverage for %s", resp, area_def)
        return {}
    return xobjs


def get_area_def_list_from_dict(area_defs):
    """Get a list of actual area_defs from full dictionary.

    Dict returned from get_area_defs_from_available_sectors
    """
    list_area_defs = []
    for area_def_id in area_defs:
        for sector_type in area_defs[area_def_id]:
            for ad in area_defs[area_def_id][sector_type]:
                list_area_defs += [area_defs[area_def_id][sector_type]["area_def"]]
    return list_area_defs


def set_comparison_path(output_dict, product_name, output_type, command_line_args=None):
    """Replace variables specified by <varname> in compare_path.

    Parameters
    ----------
    config : dict
        Dictionary of output specifications, containing key "compare_path"
    product_name : str
        Current requested product name, all instances of
        <product> in compare_path replaced with product_name argument
    output_type : str
        Current requested output type, all instances of
        <output> in compare_path replaced with output argument

    Returns
    -------
    str
        Return a single string with the fully specified comparison path for
        current product
    """
    compare_path = None
    if command_line_args is not None and command_line_args["compare_path"] is not None:
        compare_path = command_line_args["compare_path"]

    # If this config has a compare_path specified, replace variables appropriately
    if "compare_path" in output_dict or compare_path is not None:

        if compare_path is None:
            compare_path = output_dict["compare_path"]

        cpath = (
            compare_path.replace("<product>", product_name)
            .replace("<procflow>", "config_based")
            .replace("<output>", output_type)
        )
    # If there is no comparison specified, identify as "no_comparison"
    else:
        cpath = "no_comparison"

    return cpath


def initialize_final_products(final_products, cpath):
    """Initialize the final_products dictionary with cpath dict key if needed.

    Parameters
    ----------
    final_products : dict
        Dictionary of final products, with keys of final required "compare_path"
        Products with no compare_path specified are stored with the key
        "no_comparison"
    cpath : str
        Key to add to final_products dictionary

    Returns
    -------
    dict
        Return final_products dictionary, updated with current "cpath" key:
        final_products[cpath]['files'] = <list_of_files_in_given_cpath>
    """
    if cpath not in final_products:
        final_products[cpath] = {}
        # This is where we store all the files
        final_products[cpath]["files"] = []
        final_products[cpath]["database writes"] = []

    return final_products


def process_unsectored_data_outputs(
    final_products,
    available_outputs_dict,
    available_sectors_dict,
    xobjs,
    variables,
    command_line_args=None,
    write_to_product_db=False,
    config_dict=None,
):
    """Process unsectored data output.

    Loop through all possible outputs, identifying output types that require
    unsectored data output. Produce all required unsectored data output,
    update final_products dictionary accordingly, and
    return final_products dictionary with the new unsectored outputs.

    Parameters
    ----------
    final_products : dict
        Dictionary of final products, with keys of final required "compare_path"
        Products with no compare_path specified are stored with the key
        "no_comparison"
    available_outputs_dict : dict
        Dictionary of all available output product specifications
    available_sectors_dict : dict
        Dictionary of available sector types - we are looking for available
        sectors that contain the "unsectored" keyword.
    xobjs : dict
        Dictionary of xarray datasets, for use in producing unsectored output
        formats
    variables : list
        List of strings of required variables in the given product.

    Returns
    -------
    dict
        Return final_products dictionary, updated with current "cpath" key:
        final_products[cpath]['files'] = <list_of_files_in_given_cpath>
    """
    # These are the different sectors, one for each method of reprojecting or
    # sectoring or resampling the data
    for sector_type in available_sectors_dict:
        # We are looking for a sector_type that has the keyword "unsectored" meaning we
        # want to process the dat before doing anything else to it
        if (
            "unsectored" in available_sectors_dict[sector_type]
            and available_sectors_dict[sector_type]["unsectored"]
        ):
            # Once we've found an "unsectored" data type, we will look for all the
            # output_types in "available_outputs_dict" that use that sector_type
            for output_type in available_outputs_dict:
                output_dict = available_outputs_dict[output_type]
                output_dict = update_output_dict_from_command_line_args(
                    output_dict, command_line_args
                )
                if output_dict["requested_sector_type"] == sector_type:
                    # Now we will produce all of the individual products for the given
                    # output_type/sector_type
                    for product_name in output_dict["product_names"]:
                        # This grabs the compare_path that was requested in the YAML
                        # config, and replaces all instances of <product> with
                        # product_name and all instances of <output> with output_type
                        cpath = set_comparison_path(
                            output_dict, product_name, output_type, command_line_args
                        )
                        # This adds "cpath" to the final_products dictionary, if
                        # necessary
                        final_products = initialize_final_products(
                            final_products, cpath
                        )

                        # This actually produces all the required output files for the
                        # current produsct
                        prod_plugin = products.get_plugin(
                            xobjs["METADATA"].source_name,
                            product_name,
                            output_dict.get("product_spec_override"),
                        )
                        LOG.interactive(
                            "Processing 'xarray_dict_to_output_format' product '%s'",
                            product_name,
                        )
                        out = process_xarray_dict_to_output_format(
                            xobjs, variables, prod_plugin, output_dict
                        )

                        # Add them to the final_products dictionary - comparisons happen
                        # at the end.
                        final_products[cpath]["files"] += out
                        if write_to_product_db:
                            for fprod in out.keys():
                                product_added = write_to_database(
                                    fprod,
                                    product_name,
                                    xobjs,
                                    available_sectors_dict,
                                    output_dict,
                                    geoips_version,
                                    config_dict=config_dict,
                                )
                                final_products[cpath]["database writes"] += [
                                    product_added
                                ]
    return final_products


def requires_bg(available_outputs_dict, sector_type):
    """Check if current sector requires background imagery.

    Check if a given sector_type is requested for any product_types that also
    require background imagery.

    Parameters
    ----------
    available_outputs_dict : dict
        Dictionary of all requested output_types (specified in YAML config)
    sector_type : str
        sector_type to determine if any output_types that require background
        imagery also request the passed sector_type

    Returns
    -------
    bool
        * True if any output_types that require background imagery require
          the passed "sector_type"
        * False if no output_types require both background imagery and the
          passed "sector_type"
    """
    # Check each output_type in the full config_dict
    for output_type in available_outputs_dict:
        # If the current output_type has an entry for "background_products" that means
        # it requires background imagery If the current output_type also requested the
        # passed "sector_type", then return True.
        if (
            "background_products" in available_outputs_dict[output_type]
            and available_outputs_dict[output_type]["requested_sector_type"]
            == sector_type
        ):
            return True
    # If no output_types required both background_products and the passed
    # "sector_type" then return False
    return False


def is_required_sector_type(available_outputs_dict, sector_type):
    """Check if current sector is required for any outputs.

    Check if a given sector_type is required for any currently requested
    output_types

    Parameters
    ----------
    available_outputs_dict : dict
        Dictionary of all requested output_types (specified in YAML config)
    sector_type : str
        Determine if any output_types require the currently requested
        "sector_type"

    Returns
    -------
    bool
        * True if any output_types require the passed "sector_type"
        * False if no output_types require the passed "sector_type"
    """
    # Go through each output_type currently requested in the YAML config file
    for output_type in available_outputs_dict.keys():
        # If the passed sector_type is requested for any output_type in the YAML
        # config, return True
        if sector_type == available_outputs_dict[output_type]["requested_sector_type"]:
            return True
    # If the passed sector_type is not needed in the YAML config, return False
    return False


def get_config_dict(config_yaml_file):
    """Populate the full config dictionary from a given YAML config file.

    Includes both sector and output specifications.

    Parameters
    ----------
    config_yaml_file : str
        Full path to YAML config file, containing sector and output
        specifications. YAML config files support environment variables in
        entries flagged with !ENV

    Returns
    -------
    dict
        Return dictionary of both sector and output specifications,
        as found in config_yaml_file. The output dictionary references the
        "sector_types" found in the available_sectors dictionary,
        each output_type requests a specific "sector_type" to be used for
        processing.
    """
    # import yaml
    # with open(config_yaml_file, 'r') as f:
    #     config_dict = yaml.safe_load(f)
    # return config_dict
    # This allows environment variables specified by !ENV ${ENVVARNAME}
    config_dict = parse_config(config_yaml_file)
    config_dict["procflow_config_file"] = config_yaml_file

    return config_dict


def get_variables_from_available_outputs_dict(
    available_outputs_dict, source_name, sector_types=None
):
    """Get required variables for all outputs for a given "source_name".

    Outputs specified within the YAML config.

    Parameters
    ----------
    available_outputs_dict : dict
        Dictionary of all requested output_types (specified in YAML config)
    source_name : str
        Find all required variables for the passed "source_name"
    sector_types : list, default=None
        if sector_types list of strings is passed, only include
        output_types that require one of the passed "sector_types"

    Returns
    -------
    list
        List of all required variables for all output products for the given
        source_name
    """
    variables = []
    # Loop through all possible output types
    for output_type in available_outputs_dict:
        # If we requested specific sector_types, only include output_types that
        # require that sector_type
        if (
            sector_types is None
            or available_outputs_dict[output_type]["requested_sector_type"]
            in sector_types
        ):
            # Loop through all products for the given output_type
            for product_name in available_outputs_dict[output_type]["product_names"]:
                # Add all required variables for the current product and source to the
                # list
                prod_plugin = products.get_plugin(
                    source_name,
                    product_name,
                    available_outputs_dict[output_type].get("product_spec_override"),
                )
                variables += get_required_variables(prod_plugin)
    # Return list of all required variables
    return list(set(variables))


def get_area_defs_from_available_sectors(
    available_sectors_dict, command_line_args, xobjs, variables
):
    """Get all required area_defs for the given set of parameters.

    YAML config parameters (config_dict), command_line_args,
    xobjs, and required variables. Command line args override config
    specifications.

    Parameters
    ----------
    available_sectors_dict : dict
        Dictionary of all requested sector_types (specified in YAML config)
    command_line_args : dict
        Dictionary of command line arguments - any command line argument that is
        also a key in available_sectors_dict[<sector_type>] will replace the
        value in the available_sectors_dict[<sector_type>]
    xobjs : dict
        Dictionary of xarray datasets, used in determining start/end time of
        data files for identifying dynamic sectors
    variables : list
        List of required variables, for determining center coverage for TCs

    Returns
    -------
    dict
        Dictionary of required area_defs, with area_def.name as the dictionary
        keys. Based on YAML config-specified available_sectors, and command
        line args

    Notes
    -----
    * Each area_def.name key has one or more "sector_types" associated with it.
    * Each sector_type dictionary contains the actual "requested_sector_dict"
      from the YAML config, and the actual AreaDefinition object that was
      returned.

        * ``area_defs[area_def.name][sector_type]['requested_sector_dict']``
        * ``area_defs[area_def.name][sector_type]['area_def']``
    """
    area_defs = {}

    # Loop through all available sector types
    for sector_type in available_sectors_dict:
        sector_dict = available_sectors_dict[sector_type].copy()

        # If the current sector_type is "unsectored" skip it, because it has no
        # associated sector information
        if "unsectored" in sector_dict and sector_dict["unsectored"]:
            continue

        # command_line_args take priority over config args - if someone passes something
        # in explicitly, it will be used rather than config "default"
        for argname in command_line_args.keys():
            if command_line_args[argname]:
                sector_dict[argname] = command_line_args[argname]

        # Double check if tcdb should be set to false
        if sector_dict.get("trackfiles"):
            sector_dict["tcdb"] = False

        # Check if sector_list specified under YAML output config file is a list or a
        # dictionary. If sector_list is a list, static sectors are enabled for all
        # platforms that use the output config YAML. If sector_list is a dictionary,
        # each key is a platform name that holds a list of static sectors to be
        # processed for said platform. If sector_list is a dictionary, and the platform
        # name is not a key, warning is raised and sector_list is set as an empty list.
        if sector_dict.get("sector_list") and isinstance(
            sector_dict.get("sector_list"), dict
        ):
            try:
                sector_dict["sector_list"] = sector_dict["sector_list"][
                    xobjs["METADATA"].platform_name
                ]
            except KeyError as resp:
                LOG.warning("%s MISSING PLATFORM NAME", resp)
                sector_dict["sector_list"] = []

        # This is the standard "get_area_defs_from_command_line_args", YAML config
        # specified sector information matches the command line specified sector
        # information
        curr_area_defs = get_area_defs_from_command_line_args(
            sector_dict, xobjs, variables, filter_time=True
        )

        # Loop through the list of area_defs returned by
        # get_area_defs_from_command_line_args, we are going to organize them
        for area_def in curr_area_defs:
            # Use description or name so it includes synoptic time
            # We want each sectorname as a key in the dictionary, with one or more
            # sector_types attached to it. Ie, we may have different sizes/resolutions
            # for the same region, so we want a dictionary of sector_types
            # within the dictionary of area_defs
            if area_def.name not in area_defs:
                # Store the actual sector_dict and area_def in the dictionary
                area_defs[area_def.name] = {
                    sector_type: {
                        "requested_sector_dict": sector_dict,
                        "area_def": area_def,
                    }
                }
            else:
                area_defs[area_def.name][sector_type] = {
                    "requested_sector_dict": sector_dict,
                    "area_def": area_def,
                }
    return area_defs


def call(fnames, command_line_args=None):
    """Workflow for efficiently running all required outputs.

    Includes all sectors and products specified in a YAML output config file.
    Specified via a YAML config file

    Parameters
    ----------
    fnames : list
        List of strings specifying full paths to input file names to process
    command_line_args : dict
        dictionary of command line arguments

    Returns
    -------
    int
        0 for successful completion,
        non-zero for error (incorrect comparison, or failed run)
    """
    ss_pid = getpid()
    pid_track = PidLog(ss_pid, logstr="MEMUSG")

    LOG.interactive("GEOIPS_VERS {}".format(geoips_version))

    process_datetimes = {}
    process_datetimes["overall_start"] = datetime.utcnow()
    final_products = {}
    removed_products = []
    saved_products = []
    num_jobs = 0

    # These args should always be checked
    check_args = [
        "output_config",
        "reader_kwargs",
        "fuse_files",
        "fuse_reader",
        "fuse_reader_kwargs",
        "fuse_resampled_read",
        "fuse_product",
        "filename_formatter_kwargs",
        "metadata_filename_formatter_kwargs",
        "tcdb_sector_list",
        "window_start_time",
        "window_end_time",
        "product_db",
        "product_db_writer_override",
        "output_file_list_fname",
    ]

    check_command_line_args(check_args, command_line_args)
    config_dict = get_config_dict(command_line_args["output_config"])

    if not fnames and "filenames" in config_dict:
        fnames = glob(config_dict["filenames"])

    if not fnames:
        raise IOError(
            "No files found on disk. "
            "Please include valid files either at the command line, "
            "or within the YAML config"
        )
    for fname in fnames:
        if not exists(fname):
            raise IOError(
                f"File '{fname}' not found. "
                "Please include valid files either at the command line, "
                "or within the YAML config"
            )

    output_file_list_fname = command_line_args["output_file_list_fname"]
    reader_kwargs = None
    bg_files = None
    bg_product_name = None
    bg_reader_kwargs = None
    bg_resampled_read = False
    # bg_self_register_dataset = None
    # bg_self_register_source = None

    # 1. get reader_kwargs from command line if specified command line
    # 2. next get reader_kwargs from YAML config_dict if specified within YAML config
    # 3. Finally, if not specified elsewhere, default reader_kwargs to {}
    if command_line_args.get("reader_kwargs"):
        reader_kwargs = command_line_args["reader_kwargs"]
    elif "reader_kwargs" in config_dict:
        reader_kwargs = config_dict.get("reader_kwargs")
    if not reader_kwargs:
        reader_kwargs = {}
    if command_line_args.get("no_presectoring"):
        presector_data = not command_line_args["no_presectoring"]
    elif "no_presectoring" in config_dict:
        presector_data = not bool(config_dict["no_presectoring"])
    else:
        presector_data = True
    if command_line_args.get("output_checker_kwargs") is not None:
        output_checker_kwargs = command_line_args["output_checker_kwargs"]
    else:
        output_checker_kwargs = {}

    # Allow pulling command line arguments from either command line or YAML config.
    # Command line arguments override YAML config

    window_start_time = None
    window_end_time = None
    if command_line_args.get("window_start_time") is not None:
        window_start_time = command_line_args["window_start_time"]
    elif "window_start_time" in config_dict:
        window_start_time = config_dict["window_start_time"]

    if command_line_args.get("window_end_time") is not None:
        window_end_time = command_line_args["window_end_time"]
    elif "window_end_time" in config_dict:
        window_end_time = config_dict["window_end_time"]

    # Allow pulling command line arguments from either command line or YAML config.
    # Command line arguments override YAML config

    window_start_time = None
    window_end_time = None
    if command_line_args.get("window_start_time") is not None:
        window_start_time = command_line_args["window_start_time"]
    elif "window_start_time" in config_dict:
        window_start_time = config_dict["window_start_time"]

    if command_line_args.get("window_end_time") is not None:
        window_end_time = command_line_args["window_end_time"]
    elif "window_end_time" in config_dict:
        window_end_time = config_dict["window_end_time"]

    if command_line_args.get("fuse_files") is not None:
        bg_files = command_line_args["fuse_files"][0]
    elif "fuse_files" in config_dict:
        bg_files = glob(config_dict["fuse_files"])

    if command_line_args.get("fuse_reader") is not None:
        bg_reader_plugin = readers.get_plugin(command_line_args["fuse_reader"][0])
    elif "fuse_reader" in config_dict:
        bg_reader_plugin = readers.get_plugin(config_dict["fuse_reader"])

    # 1. get bg_reader_kwargs from command line if specified command line
    # 2. next get bg_reader_kwargs from YAML config_dict if specified within YAML
    # 3. Finally, if not specified elsewhere, default bg_reader_kwargs to {}
    if command_line_args.get("fuse_reader_kwargs") is not None:
        bg_reader_kwargs = command_line_args["fuse_reader_kwargs"][0]
    elif "fuse_reader_kwargs" in config_dict:
        bg_reader_kwargs = config_dict["fuse_reader_kwargs"]
    if not bg_reader_kwargs:
        bg_reader_kwargs = {}

    if command_line_args.get("fuse_product") is not None:
        bg_product_name = command_line_args["fuse_product"][0]
    elif "fuse_product" in config_dict:
        bg_product_name = config_dict["fuse_product"]

    if command_line_args.get("fuse_resampled_read") is not None:
        bg_resampled_read = command_line_args["fuse_resampled_read"][0]
    elif "fuse_resampled_read" in config_dict:
        bg_resampled_read = config_dict["fuse_resampled_read"]

    # if command_line_args.get("fuse_self_register_dataset") is not None:
    #     bg_self_register_dataset = command_line_args["fuse_self_register_dataset"][0]
    # elif "fuse_self_register_dataset" in config_dict:
    #     bg_self_register_dataset = config_dict["fuse_self_register_dataset"]

    # if command_line_args.get("fuse_self_register_source") is not None:
    #     bg_self_register_source = command_line_args["fuse_self_register_source"][0]
    # elif "fuse_self_register_source" in config_dict:
    #     bg_self_register_source = config_dict["fuse_self_register_source"]

    if command_line_args.get("product_db"):
        product_db = command_line_args["product_db"]
    elif "product_db" in config_dict:
        product_db = config_dict["product_db"]

    else:
        product_db = False

    if command_line_args.get("product_db_writer_override"):
        for sector, database_writer in command_line_args[
            "product_db_writer_override"
        ].items():
            sector_settings = config_dict["available_sectors"][sector]
            if database_writer.get("product_database_writer"):
                sector_settings["product_database_writer"] = database_writer[
                    "product_database_writer"
                ]
            if database_writer.get("product_database_writer_kwargs"):
                for key, val in database_writer[
                    "product_database_writer_kwargs"
                ].items():
                    sector_settings["product_database_writer_kwargs"][key] = val

    if command_line_args.get("composite_output_kwargs_override"):
        for sector_output, kwargs in command_line_args[
            "composite_output_kwargs_override"
        ].items():
            config_dict["outputs"][sector_output]["composite_kwargs"] = kwargs

    if bg_files is not None:
        LOG.interactive(
            "Reading background datasets using reader '%s'...", bg_reader_plugin.name
        )
        bg_xobjs = bg_reader_plugin(bg_files, metadata_only=True, **bg_reader_kwargs)
        prod_plugin = products.get_plugin(
            bg_xobjs["METADATA"].source_name,
            bg_product_name,
        )
        bg_variables = get_required_variables(prod_plugin)
        LOG.interactive("Variables '%s'...", bg_variables)

    if product_db:
        from os import getenv

        if not getenv("GEOIPS_DB_URI"):
            raise ValueError("Need to set both $GEOIPS_DB_URI")

    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
    reader_plugin = readers.get_plugin(config_dict["reader_name"])
    LOG.interactive(
        "Reading metadata from datasets using reader '%s'...", reader_plugin.name
    )
    xobjs = reader_plugin(fnames, metadata_only=True, **reader_kwargs)
    source_name = xobjs["METADATA"].source_name

    if not produce_current_time(config_dict, xobjs["METADATA"], output_dict_keys=None):
        LOG.interactive("SKIPPING ALL PROCESSING no products required for current time")
        return 0

    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
    variables = get_variables_from_available_outputs_dict(
        config_dict["outputs"], source_name
    )

    # If this config does not perform a sectored read, just read all the data now
    # Otherwise data will be read within the area_def loop
    sectored_read = False
    resampled_read = False
    if "sectored_read" in config_dict and config_dict["sectored_read"]:
        sectored_read = True
    if "resampled_read" in config_dict and config_dict["resampled_read"]:
        resampled_read = True

    if not resampled_read and not sectored_read:
        pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
        LOG.interactive("Reading full dataset using reader '%s'...", reader_plugin.name)
        xobjs = reader_plugin(
            fnames, metadata_only=False, chans=variables, **reader_kwargs
        )

    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)

    # command_line_args take priority over config args - if someone passes something in
    # explicitly, it will be used rather than config "default"
    area_defs = get_area_defs_from_available_sectors(
        config_dict["available_sectors"], command_line_args, xobjs, variables
    )

    # Check if we have any required unsectored outputs, if so produce here,
    # then continue
    LOG.interactive("\n\n\n\nNEXT Processing any unsectored data outputs...\n\n")
    final_products = process_unsectored_data_outputs(
        final_products,
        config_dict["outputs"],
        config_dict["available_sectors"],
        xobjs,
        variables,
        command_line_args,
        write_to_product_db=product_db,
        config_dict=config_dict,
    )
    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)

    list_area_defs = get_area_def_list_from_dict(area_defs)

    LOG.interactive(
        "\n\n\n\nNEXT Processing %s sectored data outputs...\n\n", len(area_defs)
    )
    area_def_num = 0
    # Loop through each template - register the data once for each template/area_def
    for area_def_id in area_defs:
        area_def_num = area_def_num + 1

        LOG.interactive(
            "\n\n\n\nNEXT area def id: %s (%s of %s)\n\n",
            area_def_id,
            area_def_num,
            len(area_defs),
        )

        bg_alg_xarrays = {}
        # Loop through each sector_type - each sector_type is a different projection /
        # shape / resolution, so we only want to reproject once for each sector_type
        sector_type_num = 0
        for sector_type in area_defs[area_def_id]:
            sector_type_num = sector_type_num + 1
            # Rather than solely relying on the metadata to obtain the source_name,
            # check the source name for all datasets. We cannot guarantee the
            # source_name in the METADATA is the end-all-be-all. Use the first product
            # plugin that we find.
            # This should really only be toggled if we expect this to be the case, or
            # maybe change source_name to source_names in the standard GeoIPS attrs.
            # If going with the toggle solution, perhaps this would be triggered if
            # METADATA.source_name == multi, or something along those lines.
            all_source_names = [x.source_name for x in xobjs.values()]
            for src_nm in set(all_source_names):
                try:
                    curr_variables = get_variables_from_available_outputs_dict(
                        config_dict["outputs"], src_nm, sector_types=[sector_type]
                    )
                    break
                except PluginError as e:
                    LOG.warn(e)
                    continue
            else:
                raise PluginError(
                    f"Plugin doesn't exist under source names" f"{all_source_names}"
                )
            if not curr_variables:
                LOG.info("No input variables for sector type: %s", sector_type)
                continue

            # If we read separately for each sector (geostationary), then must set
            # xobjs within area_def loop
            if sectored_read:
                pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
                # This will return potentially multiple sectored datasets of different
                # shapes/resolutions. Note currently get_sectored_read and
                # get_resampled_read are identical, because we have no
                # sectored_read based readers.
                LOG.interactive(
                    "Performing sectored read with reader '%s'", reader_plugin.name
                )
                xobjs = get_sectored_read(
                    config_dict,
                    area_defs,
                    area_def_id,
                    sector_type,
                    reader_plugin,
                    reader_kwargs,
                    fnames,
                    curr_variables,
                )
                if not xobjs:
                    continue
            if resampled_read:
                pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
                # This will return one resampled dataset
                # Note currently get_sectored_read and get_resampled_read are identical,
                # because we have no sectored_read based readers.
                LOG.interactive(
                    "Performing resampled read with reader '%s'", reader_plugin.name
                )
                xobjs = get_resampled_read(
                    config_dict,
                    area_defs,
                    area_def_id,
                    sector_type,
                    reader_plugin,
                    reader_kwargs,
                    fnames,
                    curr_variables,
                )
                if not xobjs:
                    continue

            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
            area_def = area_defs[area_def_id][sector_type]["area_def"]

            # Padded region to ensure we have enough data for recentering, etc.
            # Do NOT pad if we are using a reader_defined or self_register area_def -
            # that indicates we are going to use all of the data we have, so we
            # will not sector
            if area_def.sector_type not in ["reader_defined", "self_register"]:
                pad_area_def = pad_area_definition(
                    area_def, xobjs["METADATA"].source_name
                )
            else:
                pad_area_def = area_def

            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
            # See if this sector_type is used at all for product output, if not, skip
            # it.
            if not is_required_sector_type(config_dict["outputs"], sector_type):
                LOG.interactive(
                    "\n\n\nSKIPPING sector type: %s, not required for outputs %s",
                    sector_type,
                    config_dict["outputs"].keys(),
                )
                continue
            requested_sector_dict = area_defs[area_def_id][sector_type][
                "requested_sector_dict"
            ]

            LOG.interactive(
                "\n\n\n\nNEXT area def id: %s (%s of %s), "
                "sector_type: %s (%s of %s)\n\n",
                area_def_id,
                area_def_num,
                len(area_defs),
                sector_type,
                sector_type_num,
                len(area_defs[area_def_id]),
            )

            LOG.info("\n\n\n\narea definition: %s", area_def)
            LOG.info("\n\n\n\nrequested sector dict: %s\n\n\n\n", requested_sector_dict)

            # Reduce hours before and after sector time, so we don't get both overpasses
            # from a single. Sector to pad_area_def so we have enough data for
            # recentering.
            process_datetimes[area_def.area_id] = {}
            process_datetimes[area_def.area_id]["start"] = datetime.utcnow()

            # Make sure we grab some around the required data.
            # Do NOT sector if we are using a reader_defined or self_register area_def -
            # that indicates we are going to use all of the data we have, so we will not
            # sector
            if area_def.sector_type not in ["reader_defined", "self_register"]:
                if presector_data:
                    LOG.interactive("Sectoring xarrays, drop=True")
                    # window start/end time override hours before/after sector time.
                    pad_sect_xarrays = sector_xarrays(
                        xobjs,
                        pad_area_def,
                        varlist=curr_variables,
                        hours_before_sector_time=6,
                        hours_after_sector_time=9,
                        drop=True,
                        window_start_time=window_start_time,
                        window_end_time=window_end_time,
                    )
                else:
                    pad_sect_xarrays = xobjs
            else:
                pad_sect_xarrays = xobjs

            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)

            # See what variables are left after sectoring (could lose some due to
            # day/night)
            all_vars = []
            for key, xobj in pad_sect_xarrays.items():
                # Double check the xarray object actually contains data
                for var in list(xobj.variables.keys()):
                    if xobj[var].count() > 0:
                        all_vars.append(var)

            # If we didn't get any data, continue to the next sector_type
            if len(pad_sect_xarrays) == 0:
                LOG.interactive(
                    "SKIPPING no pad_area_def pad_sect_xarrays returned for %s",
                    area_def.name,
                )
                continue

            # Now we check to see if the current area_def is the closest one to
            # the dynamic time, if appropriate. We could end up with multiple
            # area_defs for a single dynamic sector, and we can't truly test to see
            # how close each one is to the actual data until we sector it...
            # So, check now to see if any of the area_defs in list_area_defs is
            # closer than pad_area_def
            if not verify_area_def(
                list_area_defs,
                pad_area_def,
                pad_sect_xarrays["METADATA"].start_datetime,
                pad_sect_xarrays["METADATA"].end_datetime,
            ):
                LOG.interactive(
                    "SKIPPING duplicate area_def, out of time range, for %s",
                    area_def.name,
                )
                continue

            # Check the config dict to see if this sector_type requests background
            # products
            if bg_files and requires_bg(config_dict["outputs"], sector_type):
                # If we haven't created the bg_alg_xarray for the current sector_type
                # yet, process it and add to the dictionary
                if sector_type not in bg_alg_xarrays:
                    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
                    bg_pad_sect_xarrays = None
                    try:
                        LOG.interactive(
                            "Reading background data with reader '%s'",
                            bg_reader_plugin.name,
                        )
                        bg_xobjs = bg_reader_plugin(
                            bg_files,
                            metadata_only=False,
                            chans=bg_variables,
                            area_def=pad_area_def,
                        )
                        if presector_data:
                            LOG.interactive(
                                "Sectoring background data, variables %s", bg_variables
                            )
                            # window start/end time override hours before/after sector
                            # time.
                            bg_pad_sect_xarrays = sector_xarrays(
                                bg_xobjs,
                                pad_area_def,
                                varlist=bg_variables,
                                hours_before_sector_time=6,
                                hours_after_sector_time=9,
                                drop=True,
                                window_start_time=window_start_time,
                                window_end_time=window_end_time,
                            )
                        else:
                            bg_pad_sect_xarrays = bg_xobjs
                    except CoverageError as resp:
                        LOG.warning(
                            f"{resp} SKIPPING - NO COVERAGE FOR BACKGROUND DATA"
                        )
                    # Only attempt to get bg xarrays if they weren't sectored away to
                    # nothing.
                    bg_prod_plugin = products.get_plugin(
                        bg_xobjs["METADATA"].source_name, bg_product_name
                    )
                    if bg_pad_sect_xarrays:
                        bg_alg_xarrays[sector_type] = get_bg_xarray(
                            bg_pad_sect_xarrays,
                            area_def,
                            bg_prod_plugin,
                            resampled_read=bg_resampled_read,
                        )
            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)

            # Must adjust the area definition AFTER sectoring xarray (to get valid
            # start/end time
            sector_adjuster = None
            if "sector_adjuster" in config_dict["available_sectors"][sector_type]:
                sector_adjuster = config_dict["available_sectors"][sector_type][
                    "sector_adjuster"
                ]

            adadj_fnames = []
            if sector_adjuster:
                LOG.info("\n\n\n\nAdjusting Area Definition: %s", sector_adjuster)
                LOG.info(
                    "\n\n\n\nBEFORE ADJUSTMENT area definition: %s\n\n\n\n", area_def
                )
                sect_adj_plugin = sector_adjusters.get_plugin(sector_adjuster)
                # Use normal size sectored xarray when running sector_adjuster, not
                # padded. Center time (mintime + (maxtime - mintime)/2) is very slightly
                # different for different size sectored arrays, so for consistency if we
                # change padding amounts, use the fully sectored array for adjusting the
                # area_def.
                if pad_sect_xarrays["METADATA"].source_name not in ["amsu-b", "mhs"]:
                    # The exact sectored arrays, without padding.
                    # Note this must be sectored both before AND after sector_adjuster -
                    # to ensure we both have an accurate center time for adjustments,
                    # and so we get all of the data.
                    if area_def.sector_type not in ["reader_defined", "self_register"]:
                        LOG.interactive(
                            "Sectoring xarrays for sector adjuster '%s'",
                            sector_adjuster,
                        )
                        if presector_data:
                            # window start/end time override hours before/after sector
                            # time.
                            sect_xarrays = sector_xarrays(
                                pad_sect_xarrays,
                                area_def,
                                varlist=curr_variables,
                                hours_before_sector_time=6,
                                hours_after_sector_time=9,
                                drop=True,
                                window_start_time=window_start_time,
                                window_end_time=window_end_time,
                            )
                        else:
                            sect_xarrays = pad_sect_xarrays
                    else:
                        sect_xarrays = pad_sect_xarrays
                    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
                    # If we didn't get any data, continue to the next sector_type
                    # Note we can have coverage for pad_sect_xarrays, but none for
                    # sect_xarrays - ensure we also skip no coverage for sect_xarrays
                    if len(sect_xarrays) == 0:
                        LOG.interactive(
                            "SKIPPING no area_def sect_xarrays returned for %s",
                            area_def.name,
                        )
                        continue
                    if (
                        sect_adj_plugin.family
                        == "list_xarray_list_variables_to_area_def_out_fnames"
                    ):
                        LOG.interactive("Adjusting sector with '%s'", sector_adjuster)
                        area_def, adadj_fnames = sect_adj_plugin(
                            list(sect_xarrays.values()),
                            area_def,
                            curr_variables,
                            config_dict["available_sectors"][sector_type][
                                "adjust_variables"
                            ],
                        )
                    else:
                        LOG.interactive("Adjusting sector with '%s'", sector_adjuster)
                        area_def = sect_adj_plugin(
                            list(sect_xarrays.values()),
                            area_def,
                            curr_variables,
                            config_dict["available_sectors"][sector_type][
                                "adjust_variables"
                            ],
                        )
                else:
                    # AMSU-b specifically needs full swath width... Need a way to
                    # generalize this.
                    if (
                        sect_adj_plugin.family
                        == "list_xarray_list_variables_to_area_def_out_fnames"
                    ):
                        LOG.interactive("Adjusting sector with '%s'", sector_adjuster)
                        area_def, adadj_fnames = sect_adj_plugin(
                            list(pad_sect_xarrays.values()),
                            area_def,
                            curr_variables,
                            config_dict["available_sectors"][sector_type][
                                "adjust_variables"
                            ],
                        )
                    else:
                        LOG.interactive("Adjusting sector with '%s'", sector_adjuster)
                        area_def = sect_adj_plugin(
                            list(pad_sect_xarrays.values()),
                            area_def,
                            curr_variables,
                            config_dict["available_sectors"][sector_type][
                                "adjust_variables"
                            ],
                        )

                cpath = set_comparison_path(
                    config_dict["available_sectors"][sector_type],
                    product_name="archer",
                    output_type="archer",
                    command_line_args=command_line_args,
                )
                final_products = initialize_final_products(final_products, cpath)
                final_products[cpath]["files"] += adadj_fnames

                LOG.info(
                    "\n\n\n\nAFTER ADJUSTMENT area definition: %s\n\n\n\n", area_def
                )

            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
            # The exact sectored arrays, without padding.
            # Note this must be sectored AFTER sector_adjuster - to ensure we get all
            # the data. Do NOT sector if we are using a reader_defined or self_register
            # area_def - that indicates we are going to use all of the data we have, so
            # we will not sector
            if area_def.sector_type not in ["reader_defined", "self_register"]:
                LOG.interactive(
                    "Sectoring self register xarrays for area_def '%s'", area_def.name
                )
                if presector_data:
                    # window start/end time override hours before/after sector time.
                    sect_xarrays = sector_xarrays(
                        pad_sect_xarrays,
                        area_def,
                        varlist=curr_variables,
                        hours_before_sector_time=6,
                        hours_after_sector_time=9,
                        drop=True,
                        window_start_time=window_start_time,
                        window_end_time=window_end_time,
                    )
                else:
                    sect_xarrays = pad_sect_xarrays
            else:
                sect_xarrays = pad_sect_xarrays

            pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
            # If we didn't get any data, continue to the next sector_type
            # Note we can have coverage for pad_sect_xarrays, but none for sect_xarrays
            # - ensure we also skip no coverage for sect_xarrays
            if len(sect_xarrays) == 0:
                LOG.interactive(
                    "SKIPPING no area_def sect_xarrays returned for %s", area_def.name
                )
                continue

            # Keep track of the applied algorithms in order to prevent redundant
            # algorithm application
            pad_alg_xarrays = {}
            alg_xarrays = {}
            output_num = 0
            required_outputs = get_required_outputs(config_dict, sector_type)
            for output_type, output_dict in required_outputs.items():
                if not produce_current_time(
                    config_dict, xobjs["METADATA"], output_dict_keys=[output_type]
                ):
                    LOG.interactive(
                        """SKIPPING PROCESSING no products required for output_type {0}
                        at current time""".format(
                            output_type,
                        )
                    )
                    continue
                output_dict = update_output_dict_from_command_line_args(
                    output_dict, command_line_args
                )

                output_num = output_num + 1

                LOG.interactive(
                    "\n\n\n\nNEXT area def id: %s (%s of %s), "
                    "sector_type: %s (%s of %s), "
                    "output_type: %s (%s of %s)\n\n",
                    area_def_id,
                    area_def_num,
                    len(area_defs),
                    sector_type,
                    sector_type_num,
                    len(area_defs[area_def_id]),
                    output_type,
                    output_num,
                    len(required_outputs.keys()),
                )

                LOG.info("\n\n\n\narea definition: %s", area_def)

                product_num = 0
                all_source_names = [x.source_name for x in pad_sect_xarrays.values()]
                for product_name in output_dict["product_names"]:
                    product_num = product_num + 1
                    for source_name in set(all_source_names):
                        # Rather than solely relying on the metadata to obtain the
                        # source_name, check the source name for all datasets.
                        # We cannot guarantee the source_name in the METADATA is the
                        # end-all-be-all. Use the first product plugin that we find.
                        # This should really only be toggled if we expect this to be
                        # the case, or maybe change source_name to source_names in the
                        # standard GeoIPS attrs. If going with the toggle solution,
                        # perhaps this would be triggered if
                        # METADATA.source_name == multi, or something along those lines
                        try:
                            prod_plugin = products.get_plugin(
                                source_name,
                                product_name,
                                output_dict.get("product_spec_override"),
                            )
                            break
                        except PluginError:
                            continue
                    else:
                        raise PluginError(
                            f"Plugin [{product_name}] doesn't exist under source names"
                            "{all_source_names}"
                        )
                    LOG.info("\n\n\n\nAll area_def_ids: %s", area_defs.keys())
                    LOG.info(
                        "\n\n\n\nAll sector_types: %s", area_defs[area_def_id].keys()
                    )
                    LOG.info(
                        "\n\n\n\nAll output_types: %s", config_dict["outputs"].keys()
                    )
                    LOG.info(
                        "\n\n\n\nAll product_names: %s", output_dict["product_names"]
                    )
                    LOG.info("\n\n\n\nCurrent area definition: %s", area_def)

                    LOG.interactive(
                        "\n\n\n\nNEXT area def id: %s (%s of %s), "
                        "sector_type: %s (%s of %s), "
                        "output_type: %s (%s of %s), "
                        "product_name: %s (%s of %s)\n\n",
                        area_def_id,
                        area_def_num,
                        len(area_defs),
                        sector_type,
                        sector_type_num,
                        len(area_defs[area_def_id]),
                        output_type,
                        output_num,
                        len(required_outputs.keys()),
                        product_name,
                        product_num,
                        len(output_dict["product_names"]),
                    )

                    LOG.info(
                        """\n\n\n\nAll current output_types for sector_type {0}:
                        {1}\n\n\n\n""".format(
                            sector_type,
                            required_outputs.keys(),
                        )
                    )

                    product_variables = get_required_variables(prod_plugin)

                    # Make sure we still have all the required variables after sectoring
                    if not set(product_variables).issubset(all_vars):
                        LOG.interactive(
                            "SKIPPING product %s missing variables %s",
                            product_name,
                            set(product_variables).difference(all_vars),
                        )
                        continue
                    cpath = set_comparison_path(
                        output_dict, product_name, output_type, command_line_args
                    )
                    final_products = initialize_final_products(final_products, cpath)

                    # Produce sectored data output
                    LOG.interactive(
                        "Processing sectored data products for product '%s'",
                        prod_plugin.name,
                    )
                    curr_output_products = process_sectored_data_output(
                        pad_sect_xarrays,
                        product_variables,
                        prod_plugin,
                        output_dict,
                        area_def=area_def,
                    )
                    # If the current product required sectored data processing, skip the
                    # rest of the loop
                    if curr_output_products:
                        final_products[cpath]["files"] += curr_output_products
                        if product_db:
                            for fprod in curr_output_products:
                                LOG.interactive(
                                    "GEOIPS_VERS writing to db {}".format(
                                        geoips_version
                                    )
                                )
                                product_added = write_to_database(
                                    fprod,
                                    product_name,
                                    pad_sect_xarrays["METADATA"],
                                    config_dict["available_sectors"],
                                    output_dict,
                                    geoips_version,
                                    area_def=area_def,
                                    config_dict=config_dict,
                                )
                                if product_added is not None:
                                    final_products[cpath]["database writes"] += [
                                        product_added
                                    ]
                        continue

                    output_formatter = get_output_formatter(output_dict)
                    output_fmt_plugin = output_formatters.get_plugin(output_formatter)

                    if output_fmt_plugin.family == "xarray_data":
                        # If we're saving out intermediate data file, write out
                        # pad_area_def.
                        if product_name not in pad_alg_xarrays:
                            pad_alg_xarrays[product_name] = get_alg_xarray(
                                pad_sect_xarrays,
                                pad_area_def,
                                prod_plugin,
                                processed_xarrays=pad_alg_xarrays,
                                resector=presector_data,
                                resampled_read=resampled_read,
                                variable_names=product_variables,
                                window_start_time=window_start_time,
                                window_end_time=window_end_time,
                            )
                        else:
                            LOG.info(
                                "  product %s already in pad_alg_xarrays", product_name
                            )
                            LOG.info(
                                "  pad_alg_xarrays datasets: %s",
                                list(pad_alg_xarrays.keys()),
                            )
                            LOG.info(
                                "  pad_alg_xarrays[%s] datasets: %s",
                                product_name,
                                list(pad_alg_xarrays[product_name].variables.keys()),
                            )
                        alg_xarray = pad_alg_xarrays[product_name]
                    elif area_def.sector_type in ["reader_defined", "self_register"]:
                        alg_xarray = get_alg_xarray(
                            pad_sect_xarrays,
                            pad_area_def,
                            prod_plugin,
                            resector=False,
                            resampled_read=resampled_read,
                            variable_names=product_variables,
                            window_start_time=window_start_time,
                            window_end_time=window_end_time,
                        )
                    else:
                        # If we're writing out an image, cut it down to the desired
                        # size.
                        if product_name not in alg_xarrays:
                            alg_xarrays[product_name] = get_alg_xarray(
                                sect_xarrays,
                                area_def,
                                prod_plugin,
                                processed_xarrays=alg_xarrays,
                                resector=presector_data,
                                resampled_read=resampled_read,
                                variable_names=product_variables,
                                window_start_time=window_start_time,
                                window_end_time=window_end_time,
                            )
                            # This is a unique identifier so we can re-use variables
                            # and product arrays appropriately.
                            area_def_key = get_unique_dataset_key(
                                area_def, alg_xarrays[product_name]
                            )
                            alg_xarrays[area_def_key] = alg_xarrays[product_name]
                        else:
                            LOG.info(
                                "  product %s already in alg_xarrays", product_name
                            )
                            LOG.info(
                                "  alg_xarrays datasets: %s", list(alg_xarrays.keys())
                            )
                            LOG.info(
                                "  alg_xarrays[%s] datasets: %s",
                                product_name,
                                list(alg_xarrays[product_name].variables.keys()),
                            )
                        alg_xarray = alg_xarrays[product_name]

                    covg_plugin = get_covg_from_product(
                        prod_plugin,
                        covg_field="image_production_coverage_checker",
                    )
                    covg_args = get_covg_args_from_product(
                        prod_plugin,
                        covg_field="image_production_coverage_checker",
                    )

                    # Set variable_name to prod_plugin.name if not defined.
                    # Always use variable_name if it is defined.
                    # Remove from args so there is not a duplicate are when passing
                    # (since we are passing covg_varname explicitly).
                    # Note get_covg_args_from_product was updated to return a copy of
                    # covg_args, so this does not impact future uses of the product.
                    covg_varname = covg_args.pop("variable_name", prod_plugin.name)
                    # Note variables can be specified as DATASET:VARIABLE,
                    # since this is a preprocessed alg_xarray, and not a
                    # dictionary of datasets, just use the variable name
                    # (we expect the correct variable will exist in this
                    # final processed array)
                    if ":" in covg_varname:
                        covg_varname = covg_varname.split(":")[1]
                    covg = covg_plugin(alg_xarray, covg_varname, area_def, **covg_args)

                    fname_covg_plugin = get_covg_from_product(
                        prod_plugin,
                        covg_field="filename_coverage_checker",
                    )
                    fname_covg_args = get_covg_args_from_product(
                        prod_plugin,
                        covg_field="filename_coverage_checker",
                    )

                    # Set variable_name to prod_plugin.name if not defined.
                    # Always use variable_name if it is defined.
                    # Remove from args so there is not a duplicate are when passing
                    # (since we are passing covg_varname explicitly).
                    # Note get_covg_args_from_product was updated to return a copy of
                    # covg_args, so this does not impact future uses of the product.
                    covg_varname = fname_covg_args.pop(
                        "variable_name", prod_plugin.name
                    )
                    # Note variables can be specified as DATASET:VARIABLE,
                    # since this is a preprocessed alg_xarray, and not a
                    # dictionary of datasets, just use the variable name
                    # (we expect the correct variable will exist in this
                    # final processed array)
                    if ":" in covg_varname:
                        covg_varname = covg_varname.split(":")[1]
                    fname_covg = fname_covg_plugin(
                        alg_xarray, covg_varname, area_def, **fname_covg_args
                    )

                    minimum_coverage = 10
                    config_minimum_coverage = get_minimum_coverage(
                        prod_plugin.name, output_dict
                    )
                    if hasattr(alg_xarray, "minimum_coverage"):
                        minimum_coverage = alg_xarray.minimum_coverage
                    if config_minimum_coverage is not None:
                        minimum_coverage = config_minimum_coverage
                    LOG.interactive(
                        "Required coverage %s for product %s, actual coverage %s "
                        "using %s checker, filename covg %s using %s",
                        minimum_coverage,
                        product_name,
                        covg,
                        covg_plugin.name,
                        fname_covg,
                        fname_covg_plugin.name,
                    )
                    if covg < minimum_coverage and fname_covg < minimum_coverage:
                        LOG.interactive(
                            "Insufficient coverage %s / %s for data products, SKIPPING",
                            covg,
                            fname_covg,
                        )
                        continue
                    composite_kwargs = output_dict.get("composite_kwargs", {})
                    if composite_kwargs.get("composite_products"):
                        if not product_db:
                            LOG.interactive(
                                "Product database disabled, cannot create composite"
                            )
                            continue
                        from geoips.utils.composite import find_preproc_alg_files

                        # Required kwargs for generating composite
                        comp_settings = output_dict["composite_kwargs"]
                        reader = readers.get_plugin(
                            comp_settings["composite_input_file_reader"]
                        )
                        comp_file_format = comp_settings["composite_input_file_format"]
                        composite_window = comp_settings["composite_window"]

                        # Optional kwargs for generating composite
                        db_query_plugin = comp_settings.get("database_query_module")
                        db_kwargs = config_dict["available_sectors"][sector_type].get(
                            "product_database_writer_kwargs", {}
                        )
                        # Default schema and tables
                        db_schema = db_kwargs.get("schema_name")
                        db_table = db_kwargs.get("table_name")
                        # Check to see if products should be queried from other schema
                        # and/or tables - fall back to defaults if not specified.
                        query_schema = comp_settings.get(
                            "database_query_schema", db_schema
                        )
                        query_table = comp_settings.get(
                            "database_query_table", db_table
                        )

                        preproc_files = find_preproc_alg_files(
                            product_time=alg_xarray.start_datetime,
                            composite_window=composite_window,
                            sector_name=area_def.area_id,
                            product=product_name,
                            sensor=alg_xarray.source_name,
                            platform=alg_xarray.platform_name,
                            file_format=comp_file_format,
                            product_db=product_db,
                            db_query_plugin=db_query_plugin,
                            db_schemas=[query_schema],
                            db_tables=[query_table],
                        )
                        if preproc_files:
                            pre_proc = reader(preproc_files)
                            if (
                                "rgb"
                                in prod_plugin["spec"]["colormapper"]["plugin"]["name"]
                            ):
                                rgb_var = prod_plugin.name
                            else:
                                rgb_var = None
                            alg_xarray = combine_preproc_xarrays_with_alg_xarray(
                                pre_proc, alg_xarray, rgb_var=rgb_var
                            )
                            comp_covg = covg_plugin(
                                alg_xarray, covg_varname, area_def, **covg_args
                            )
                            LOG.info("Composite coverage: %s", comp_covg)
                        else:
                            LOG.info("No files to create composite!")

                    plot_data_kwargs = get_output_formatter_kwargs(
                        output_dict,
                        alg_xarray,
                        area_def,
                        sector_type,
                        bg_files,
                        bg_alg_xarrays,
                        bg_product_name,
                    )

                    if (
                        bg_files
                        and "background_products" in output_dict
                        and sector_type in bg_alg_xarrays
                    ):
                        alg_xarray = combine_filename_extra_fields(
                            bg_alg_xarrays[sector_type], alg_xarray
                        )

                    curr_products = plot_data(
                        output_dict,
                        alg_xarray,
                        area_def,
                        prod_plugin,
                        plot_data_kwargs,
                    )

                    if curr_products:
                        final_products[cpath]["files"] += list(curr_products.keys())
                    else:
                        continue

                    if product_db:
                        for fprod in curr_products.keys():
                            product_added = write_to_database(
                                fprod,
                                product_name,
                                alg_xarray,
                                config_dict["available_sectors"],
                                output_dict,
                                geoips_version,
                                coverage=covg,
                                area_def=area_def,
                                config_dict=config_dict,
                            )
                            if product_added is not None:
                                final_products[cpath]["database writes"] += [
                                    product_added
                                ]

                    if (
                        "remove_duplicates" in output_dict
                        and output_dict["remove_duplicates"] is not None
                    ):
                        curr_removed_products, curr_saved_products = remove_duplicates(
                            curr_products, remove_files=True
                        )
                        removed_products += curr_removed_products
                        saved_products += curr_saved_products

                    process_datetimes[area_def.area_id]["end"] = datetime.utcnow()
                    num_jobs += 1

    pid_track.print_mem_usg(logstr="MEMUSG", verbose=False)
    process_datetimes["overall_end"] = datetime.utcnow()

    LOG.interactive("\n\n\nProcessing complete! Checking outputs...\n\n\n")

    retval = 0
    failed_compares = {}
    for cpath in final_products:
        if cpath != "no_comparison":
            from geoips.interfaces.module_based.output_checkers import output_checkers

            for output_product in final_products[cpath]["files"]:
                plugin_name = output_checkers.identify_checker(output_product)
                output_checker = output_checkers.get_plugin(plugin_name)
                kwargs = {}
                if output_checker.name in output_checker_kwargs:
                    kwargs = output_checker_kwargs[output_checker.name]
                curr_retval = output_checker(
                    output_checker,
                    cpath,
                    [output_product],
                    **kwargs,
                )
                retval += curr_retval
                if curr_retval != 0:
                    failed_compares[cpath] = curr_retval
        else:
            LOG.info("No comparison specified, not attempting to compare outputs")

    successful_comparison_dirs = 0
    failed_comparison_dirs = 0

    LOG.interactive(
        "\n\n\nThe following products were produced from procflow %s\n\n",
        basename(__file__),
    )
    for cpath in final_products:
        if cpath in failed_compares:
            LOG.interactive(
                "%s FAILED COMPARISONS IN DIR: %s\n", failed_compares[cpath], cpath
            )
            failed_comparison_dirs = failed_comparison_dirs + 1
        elif cpath != "no_comparison":
            LOG.info("SUCCESSFUL COMPARISON DIR: %s\n", cpath)
            successful_comparison_dirs = successful_comparison_dirs + 1
        for filename in final_products[cpath]["files"]:
            LOG.interactive(
                "    \u001b[34mCONFIGSUCCESS\033[0m %s",
                replace_geoips_paths(filename, curly_braces=True),
            )
            if filename in final_products[cpath]["database writes"]:
                LOG.interactive("    DATABASESUCCESS %s", filename)
        LOG.info("\n")

    for removed_product in removed_products:
        LOG.interactive("    DELETEDPRODUCT %s", removed_product)
        if product_db:
            LOG.info("    FLAGGING as deleted in product database")
            flag_product_as_deleted(removed_product, area_defs)

    if output_file_list_fname:
        LOG.info("Writing successful outputs to %s", output_file_list_fname)
        with open(output_file_list_fname, "w", encoding="utf8") as fobj:
            for cpath in final_products:
                LOG.info("Trying %s", cpath)
                if len(final_products[cpath]["files"]) > 0:
                    LOG.info(
                        "  WRITING %s to output file list, %s products generated",
                        cpath,
                        len(final_products[cpath]["files"]),
                    )
                    fobj.writelines(
                        "\n".join(
                            [
                                replace_geoips_paths(fname)
                                for fname in final_products[cpath]["files"]
                            ]
                        )
                    )
                    # If we don't write out the last newline, then wc won't return the
                    # appropriate number, and we won't get to the last file when
                    # attempting to loop through
                    fobj.writelines(["\n"])
                else:
                    LOG.info(
                        """  SKIPPING WRITING {0} to output file list, no products
                        generated""".format(
                            cpath,
                        )
                    )

    mem_usage_stats = pid_track.print_mem_usg(logstr="MEMUSG", verbose=True)
    LOG.interactive("READER_NAME: %s", config_dict["reader_name"])
    num_products = sum(
        [len(final_products[cpath]["files"]) for cpath in final_products]
    )
    LOG.interactive("NUM_PRODUCTS: %s", num_products)
    if product_db:
        LOG.info(
            "NUM_DATABASE_WRITES: %s",
            sum(
                [
                    len(final_products[cpath]["database writes"])
                    for cpath in final_products
                ]
            ),
        )
    LOG.interactive("NUM_DELETED_PRODUCTS: %s", len(removed_products))
    LOG.interactive("NUM_SUCCESSFUL_COMPARISON_DIRS: %s", successful_comparison_dirs)
    LOG.interactive("NUM_FAILED_COMPARISON_DIRS: %s", failed_comparison_dirs)
    output_process_times(process_datetimes, num_jobs)
    LOG.interactive("GEOIPS_VERS {}".format(geoips_version))
    if product_db:
        all_sectors_use_tcdb = all(
            [
                config_dict["available_sectors"][x].get("tcdb")
                for x in config_dict["available_sectors"].keys()
            ]
        )
        if (
            command_line_args.get("tcdb")
            or command_line_args.get("trackfiles")
            or all_sectors_use_tcdb
        ):
            sector_type = "dynamic_tc"
        else:
            sector_type = "static"

        write_stats_to_database(
            procflow_name="config_based",
            platform=xobjs["METADATA"].platform_name.lower(),
            geoips_vers=geoips_version,
            source=xobjs["METADATA"].source_name,
            product="multi",
            sector_type=sector_type,
            process_times=process_datetimes,
            num_products_created=num_products,
            num_products_deleted=len(removed_products),
            resource_usage_dict=mem_usage_stats,
        )
    else:
        LOG.interactive("NO PRODDB GEOIPS_VERS {}".format(geoips_version))
    return retval



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `describe family` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

import pytest

from geoips import interfaces
from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsDescribeFamily(BaseCliTest):
    """Unit Testing Class for Describe Family Command."""

    @property
    def command_combinations(self):
        """A list of every possible call signature for the GeoipsDescribeFamily command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            self._cmd_list = []
            base_args = ["geoips", "describe"]
            # add each family argument from every interface to the command arg list
            for interface_name in interfaces.__all__:
                interface = getattr(interfaces, interface_name)
                interface_name = interface_name.replace("_", "-")
                for alias in self.alias_mapping[interface_name] + [interface_name]:
                    for family_name in interface.supported_families:
                        self._cmd_list.append(base_args + [alias, "fam", family_name])
                        self._cmd_list.append(
                            base_args + [alias, "family", family_name]
                        )
            # Add argument list to retrieve help message
            self._cmd_list.append(base_args + ["-h"])
            # Add argument list with non_existent_interface
            self._cmd_list.append(
                base_args + ["non_existent_interface", "fam", "standard"]
            )
            self._cmd_list.append(
                base_args + ["non_existent_interface", "family", "standard"]
            )
            # Add argument list with non_existent_family
            self._cmd_list.append(
                base_args + ["algorithms", "fam", "non_existent_family"]
            )
            self._cmd_list.append(
                base_args + ["algorithms", "family", "non_existent_family"]
            )
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that `geoips describe <interface_name> family ...` error is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        # An error occurred using args. Assert that args is not valid and check the
        # output of the error.
        err_str = "usage: To use, type `geoips describe <interface_name> <sub-cmd> ...`"
        assert err_str in error

    def check_output(self, args, output):
        """Ensure that 'geoips describe <interface_name> family ...' success is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        # The args provided are valid, so test that the output is actually correct
        if "-h" in args:
            usg_str = (
                "usage: To use, type `geoips describe <interface_name> <sub-cmd> ...`"
            )
            assert usg_str in output
        else:
            # Checking that output from geoips describe plugin command is valid
            expected_outputs = [
                "Docstring",
                "Family Name",
                "Family Path",
                "Interface Name",
                "Interface Type",
                "Required Args / Schema",
            ]
            for output_item in expected_outputs:
                assert f"{output_item}:" in output


test_sub_cmd = TestGeoipsDescribeFamily()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips describe <interface_name> family <family_name> ...' commands.

    This test covers every valid combination of commands for the
    'geoips describe <interface_name> family <family_name>' command. We also test
    invalid commands, to ensure that the proper help documentation is provided for those
    using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with ie:
        - ['geoips', 'describe', <interface_name>, 'family', <family_name>])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



"""pytest test‑suite for ``lexeme.py``."""

from __future__ import annotations

import json
from typing import Dict, Set, Tuple

import pytest
from pydantic import BaseModel

from geoips.utils.types.partial_lexeme import Lexeme

REGULAR_PAIR: Tuple[str, str] = ("reader", "readers")
IRREGULAR_PAIRS = [
    ("analysis", "analyses"),
    ("index", "indices"),
    ("child", "children"),
]


def test_creation() -> None:
    """
    Test creation of a Lexeme instance from a valid string.

    Ensures that initializing Lexeme with a simple word does not raise an exception.
    """
    Lexeme("test")


def test_regular_equality() -> None:
    """
    Test equality behavior for regular singular and plural forms.

    Checks that singular and plural forms with a standard 's' suffix compare equal
    and maintain proper equality ordering (i.e., plural != singular).
    """
    singular, plural = REGULAR_PAIR
    assert Lexeme(singular) == Lexeme(plural) == singular
    assert not plural == singular


@pytest.mark.parametrize("singular,plural", IRREGULAR_PAIRS)
def test_irregular_equality(singular: str, plural: str) -> None:
    """
    Test equality behavior for irregular plural forms.

    Parameters
    ----------
    singular : str
        The singular form of an irregular noun.
    plural : str
        The plural form of the same noun.
    """
    assert Lexeme(singular) == plural
    assert Lexeme(plural) == singular


def test_hash_equivalence() -> None:
    """
    Test that hash-based structures treat singular and plural as equivalent keys.

    Notes
    -----
    Variables must be cast as a Lexeme to index correctly.
    """
    d: Dict[Lexeme, int] = {Lexeme("analysis"): 1}
    d[Lexeme("analyses")] = d.get(Lexeme("analyses"), 0) + 1
    assert len(d) == 1
    assert d["analysis"] == 2

    s: Set[Lexeme] = {Lexeme("READER")}
    s.add(Lexeme("readers"))
    assert len(s) == 1

    assert "test" in [Lexeme("test")]


def test_case_insensitivity() -> None:
    """
    Test that lexeme comparisons are case-insensitive.

    Verifies that different casing variants of the same word are considered equal.
    """
    assert Lexeme("ReadErS") == "readers"
    assert Lexeme("ReaDers") == "READERS"


def test_whitespace_trimmed() -> None:
    """
    Test that leading and trailing whitespace is trimmed.

    Ensures that spaces around the word do not affect equality.
    """
    assert Lexeme(" reader ") == "readers"


def test_ss_suffix_not_singularised() -> None:
    """
    Test that words ending with 'ss' are not truncated to singular.

    Words like 'glass' should not lose one 's' when singularized.
    """
    assert Lexeme("glass") == "glass"
    assert Lexeme("glass") != "glas"


def test_short_word_not_truncated() -> None:
    """
    Test that short words less than 3 characters are not truncated.

    Ensures that 'bus' remains unchanged and does not become 'bu'.
    """
    assert Lexeme("bus") == "bus"
    assert Lexeme("bus") != "bu"


def test_plural_property_idempotent() -> None:
    """
    Test that the plural property is idempotent for already plural words.

    When requesting the plural of a word that is already plural,
    it should remain the same.
    """
    word = Lexeme("boxes")
    assert word.plural == "boxes"


def test_additional_plural_rules() -> None:
    """
    Test additional pluralization rules for words ending in 'ch', 'sh', and 'x'.

    'dish' -> 'dishes', 'church' -> 'churches', 'box'.plural -> 'boxes'.
    """
    assert Lexeme("dish") == "dishes"
    assert Lexeme("church") == "churches"
    assert Lexeme("box").plural == "boxes"


def test_type_error_on_non_string() -> None:
    """
    Test that initializing Lexeme with a non-string raises a TypeError.

    Passing an integer should not be allowed.
    """
    with pytest.raises(TypeError):
        Lexeme(123)


def test_repr_roundtrip() -> None:
    """
    Test that repr and eval round-trip produce equivalent Lexeme objects.

    Ensures that repr(obj) yields a string that eval can recreate the same object.
    """
    obj = Lexeme("reader")
    string_repr = "Lexeme('reader')"
    assert repr(obj) == string_repr


def test_equality_with_unrelated_type() -> None:
    """
    Test that comparing Lexeme to an unrelated type returns False.

    Ensures safe equality checks with non-string types.
    """
    assert (Lexeme("reader") == 5) is False


def test_pydantic_model_serialisation() -> None:
    """
    Test Pydantic v2 integration for Lexeme fields in models.

    - Internal equality: singular/plural mapping within the model
    - Serialization: model_dump_json yields plain strings
    """

    class Plugin(BaseModel):
        kind: Lexeme
        interface: Lexeme

    model = Plugin(kind="reader", interface="readers")

    # Internal equality
    assert model.kind == model.interface == "reader"

    # Serialisation should yield plain strings
    data = json.loads(model.model_dump_json())
    assert data == {"kind": "reader", "interface": "readers"}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read AMSU-B and MHS passive microwave data files.

This reader is desgined for importing NOAA Advanced Microwave Sounding Unit
(AMSU)-B/Microwave Humidity Sounder (HMS) and EUMETSAQT MHS data files in hdf
format, such as

* NPR.MHOP.NP.D20153.S2046.E2230.B5832021.NS (N19),
* NPR.MHOP.NN.D20153.S1927.E2105.B7748081.NS (N18),
* NPR.MHOP.M1.D20153.S2229.E2318.B3998282.NS (METOP).

 V1.0:  Initial version, NRL-MRY, June 1, 2020

Basic information on AMSU-B product file::

    Input SD Variables
    (nscan, npix):
        npix=90 pixels per scan;
        nscan: vary with orbit
    chan-1 AT:  89 GHz as ch16 anttenna temperature at V-pol   FOV 16km at nadir
    chan-2 AT:  150 (157) GHz as ch17 the number in bracket is for MHS from
                                      metops at V-pol, 16km at nadir
    chan-3 AT:  183.31 +-1 GHz as ch18    at H-pol, 16km
    chan-4 AT:  183.31 +-3 GHz as ch19    at H-pol, 16km
    chan-5 AT:  183.31 +-7 (190.3) GHz as ch20    at V-pol, 16km
    lat:     -90, 90     deg
    lon:     -180, 180   deg
    RR:  surface rainrate (mm/hr)
    Snow: surafce snow cover
    IWP:  ice water path (unit?)
    SWE:  snow water equvelent  (unit)
    Sfc_type:  surface type
    Orbit_mode:   -1: ascending, 1: decending, 2: both
    SFR:  snowfall rate (unit?)
    LZ_angle:  local zinath angle (deg)
    SZ_angle:  solar zinath angle (deg)

    Vdata info (definition of AMSU-B date):
    ScanTime_year
    ScanTime_month
    ScanTime_day
    ScanTime_hour
    ScanTime_minute
    ScanTime_second
    ScanTime_Jday
    Time
"""
# Python Standard Libraries
from datetime import datetime
import logging
import os

# Third-Party Libraries
import matplotlib
import numpy as np
import pandas as pd

# library for hdf files
from pyhdf.SD import SD, SDC
from pyhdf.VS import HC

# from pyhdf.VS import *
from pyhdf.HDF import HDF
import xarray as xr

matplotlib.use("agg")

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "amsub_hdf"
source_names = ["amsu-b"]


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read AMSU-B hdf data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # fname='NPR.MHOP.NP.D20154.S1406.E1553.B5833031.NS'
    fname = fnames[0]

    LOG.info("Reading file %s", fname)

    #     check for right input AMSU-B/MHS data file

    data_name = os.path.basename(fname).split("_")[-1].split(".")[-1]

    if data_name != "NS":
        LOG.info("Warning: wrong AMSU-B/MHS data type:  data_type=", data_name)
        raise

    if ("NRP" and "MHOP") in os.path.basename(fname):
        LOG.info("found a AMSU-B/MHS hdf file")
    else:
        LOG.info("not a AMSU-B/MHS hdf file: skip it")
        raise

    """    ------  Notes  ------
       Read AMSU-B hdf files for 5 chan antenna temperature (AT) and asscoaited EDRs
         Then, transform these ATs and fields into xarray framework for GEOIPS
         ( AT will be corrected into brightness temperature (TB) later)

       Input Parameters:
           fname (str): input file name.  require a full path
       Returns:
           xarray.Dataset with required Variables and Attributes:
               Variables:
                        AMSUB vars:
                          'latitude', 'longitude', 'Ch1', 'Ch2', 'Ch3', 'Ch4','Ch4',
                          'RR', 'Snow','SWE','IWP','SFR' 'sfcType', 'time_scan'
               Attibutes:
                        'source_name', 'platform_name', 'data_provider',
                        'interpolation_radius_of_influence', 'start_datetime',
                        'end_datetime'
    """

    SData_ID = SD(fname, SDC.READ)
    VData_ID = HDF(fname, HC.READ).vstart()

    # get scan time info
    year = VData_ID.attach("ScanTime_year")[:]
    # month = VData_ID.attach("ScanTime_month")[:]
    # day = VData_ID.attach("ScanTime_dom")[:]
    hour = VData_ID.attach("ScanTime_hour")[:]
    minute = VData_ID.attach("ScanTime_minute")[:]
    # second = VData_ID.attach("ScanTime_second")[:]
    jday = VData_ID.attach("ScanTime_doy")[:]

    # get EDRs
    lat = SData_ID.select("Latitude").get()
    lon = SData_ID.select("Longitude").get()
    RR = SData_ID.select("RR").get()
    Snow = SData_ID.select("Snow").get()
    IWP = SData_ID.select("IWP").get()
    SWE = SData_ID.select("SWE").get()
    SFR = SData_ID.select("SFR").get()
    Sfc_type = SData_ID.select("Sfc_type").get()
    Chan1_AT = SData_ID.select("Chan1_AT").get()
    Chan2_AT = SData_ID.select("Chan2_AT").get()
    Chan3_AT = SData_ID.select("Chan3_AT").get()
    Chan4_AT = SData_ID.select("Chan4_AT").get()
    Chan5_AT = SData_ID.select("Chan5_AT").get()
    # SZ_angle = SData_ID.select("SZ_angle").get()
    # LZ_angle = SData_ID.select("LZ_angle").get()

    # Min, Max, and scale factors (using hdfview)
    # AT_min = 75.0
    # AT_max = 325.0
    # RR_min = 0.0
    # RR_max = 30.0
    # Snow_min = 0.0
    # Snow_max = 100.0
    # IWP_min = 0.0
    # IWP_max = 3.0
    # SWE_min = 0.0
    # SWE_max = 30.0
    # SFR_min = 0.03
    # SFR_max = 5.0

    AT_scale = 100.0
    RR_scale = 10.0
    Snow_scale = 1.0
    IWP_scale = 100.0
    SWE_scale = 100.0
    SFR_scale = 100.0

    # Adjust the values with scale factors

    RR = RR * RR_scale
    Snow = Snow / Snow_scale
    IWP = IWP / IWP_scale
    SWE = SWE / SWE_scale
    SFR = SFR / SFR_scale
    Chan1_AT = Chan1_AT / AT_scale
    Chan2_AT = Chan2_AT / AT_scale
    Chan3_AT = Chan3_AT / AT_scale
    Chan4_AT = Chan4_AT / AT_scale
    Chan5_AT = Chan5_AT / AT_scale

    SData_ID.end()  # close input file
    VData_ID.end()  # close input file

    #  -------- Apply the GEOIPS framework in XARRAY data frame ----------

    LOG.info("Making full dataframe")

    # setup the time in datetime64 format
    npix = RR.shape[1]  # pixels per scan
    nscan = RR.shape[0]  # total scans of this file

    time_scan = np.zeros((nscan, npix))

    for i in range(nscan):
        time_scan[i:] = "%04d%03d%02d%02d" % (
            year[i][0],
            jday[i][0],
            hour[i][0],
            minute[i][0],
        )

    #          ------  setup xarray variables   ------

    # namelist_amsub  = ['latitude', 'longitude', 'Chan1_AT', 'Chan2_AT', 'Chan3_AT',
    #                    'Chan4_AT','Chan5_AT', 'RR','Snow','IWP','SWE','SFR',
    #                    'Sfc_type','time']

    # setup amsub xarray
    xarray_amsub = xr.Dataset()
    xarray_amsub["latitude"] = xr.DataArray(lat)
    xarray_amsub["longitude"] = xr.DataArray(lon)
    xarray_amsub["Chan1_AT"] = xr.DataArray(Chan1_AT)
    xarray_amsub["Chan2_AT"] = xr.DataArray(Chan2_AT)
    xarray_amsub["Chan3_AT"] = xr.DataArray(Chan3_AT)
    xarray_amsub["Chan4_AT"] = xr.DataArray(Chan4_AT)
    xarray_amsub["Chan5_AT"] = xr.DataArray(Chan5_AT)
    xarray_amsub["RR"] = xr.DataArray(RR)
    xarray_amsub["Snow"] = xr.DataArray(Snow)
    xarray_amsub["IWP"] = xr.DataArray(IWP)
    xarray_amsub["SWE"] = xr.DataArray(SWE)
    xarray_amsub["SFR"] = xr.DataArray(SFR)
    xarray_amsub["sfcType"] = xr.DataArray(Sfc_type)
    xarray_amsub["time"] = xr.DataArray(
        pd.DataFrame(time_scan).astype(int).apply(pd.to_datetime, format="%Y%j%H%M")
    )

    # setup attributes

    # satID and start_end time from input filename (M3  --> METOP-3 ???)
    sat_id = os.path.basename(fname).split(".")[2]
    if sat_id == "NP":
        satid = "noaa-19"
    elif sat_id == "NN":
        satid = "noaa-18"
    elif sat_id == "M1":
        satid = "metop-b"
    elif sat_id == "M2":
        satid = "metop-a"
    elif sat_id == "M3":
        satid = "metop-c"

    yr = os.path.basename(fname).split(".")[3][1:3]
    jd = os.path.basename(fname).split(".")[3][3:6]
    hr = os.path.basename(fname).split(".")[4][1:3]
    mi = os.path.basename(fname).split(".")[4][3:6]
    start_time = "%02d%03d%02d%02d" % (int(yr), int(jd), int(hr), int(mi))
    hr = os.path.basename(fname).split(".")[5][1:3]
    mi = os.path.basename(fname).split(".")[5][3:6]
    end_time = "%02d%03d%02d%02d" % (int(yr), int(jd), int(hr), int(mi))

    # add attributes to xarray
    xarray_amsub.attrs["start_datetime"] = datetime.strptime(start_time, "%y%j%H%M")
    xarray_amsub.attrs["end_datetime"] = datetime.strptime(end_time, "%y%j%H%M")
    xarray_amsub.attrs["source_name"] = "amsu-b"
    xarray_amsub.attrs["platform_name"] = satid
    xarray_amsub.attrs["data_provider"] = "nesdis"

    # MTIFs need to be "prettier" for PMW products, so 2km resolution for final image
    # xarray_amsub.attrs['sample_distance_km'] = 15
    xarray_amsub.attrs["sample_distance_km"] = 2
    xarray_amsub.attrs["interpolation_radius_of_influence"] = 30000

    return {"AMSUB": xarray_amsub, "METADATA": xarray_amsub[[]]}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""VIIRS Level 2 NetCDF reader."""

import logging
from datetime import datetime
import os

from collections import OrderedDict
import numpy as np
import xarray

LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "viirs_l2_netcdf"


def merge_xarrays(src_xobj, dst_xobj):
    """Merge one xarray into another.

    Parameters
    ----------
    src_xobj : xarray.Dataset
        Input dataset to merge
    dst_xobj : xarray.Dataset
        Dataset where src_xobj will be merged

    Returns
    -------
    xarray.Dataset
        Merged xarray dataset
    """
    tmp_xobj = xarray.Dataset()
    for dvar in src_xobj.data_vars.keys():
        dset = src_xobj[dvar]
        if dvar not in dst_xobj:
            dst_xobj[dvar] = dset
        else:
            tmp_xobj[dvar] = xarray.concat(
                [dst_xobj[dvar], dset], dim=dst_xobj[dvar].dims[0]
            )
            dst_xobj = dst_xobj.drop(dvar)
    try:
        merged_xobj = xarray.merge([tmp_xobj, dst_xobj])
    except ValueError:
        tmp_dims = {x: tmp_xobj[x].size for x in tmp_xobj.dims}
        dst_dims = {x: dst_xobj[x].size for x in dst_xobj.dims}
        mismatching_dims = [x for x in dst_dims if tmp_dims[x] != dst_dims[x]]
        merged_xobj = xarray.combine_nested(
            [tmp_xobj, dst_xobj], concat_dim=mismatching_dims
        )
    return merged_xobj


def get_geoips_platform_name(meta_platform_name):
    """Map source file platform name to standard GeoIPS names.

    Parameters
    ----------
    meta_platform_name : str
        Name of platform in source file

    Returns
    -------
    str
        Standard GeoIPS platform name
    """
    geoips_pform_name = {"NPP": "npp", "NPP_OPS": "npp", "JPSS": "noaa20"}
    return geoips_pform_name[meta_platform_name]


def pair_file_names(file_list):
    """Attempt to group VIIRS NASA LANCE/LAADS files together by scan time.

    Parameters
    ----------
    file_list : list
        list of file names passed to reader

    Returns
    -------
    dict
        Dictionary of scan times with each key holding list of files per time.
    """
    basenames = [os.path.basename(x) for x in file_list]
    gran_times = [".".join(x.split(".")[1:3]) for x in basenames]
    unique_gran_times = sorted(np.unique(gran_times))
    paired_granules = OrderedDict()
    for gtimes in unique_gran_times:
        paired_granules[gtimes] = sorted([x for x in file_list if gtimes in x])
    return paired_granules


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read VIIRS L2 netCDF data.

    Parameters
    ----------
    fci_files : list
        list of files for a given scan time
    metadata_only : bool, optional
        Only return metadata, by default False
    chans : list, optional
        List of channels/variables to load, by default None
    area_def : pyresample area definition, optional
        Read in data for specific area definition, by default None
        (currently unsupported)
    self_register : bool, optional
        Self register data to given resolution, by default False
        (currently unsupported)

    Returns
    -------
    dict
        dictionary of xarray datasets

    Raises
    ------
    AttributeError
        Encountered unknown platform name in source file
    """
    xobj = xarray.Dataset()
    # for fname in pair_file_names(fnames):
    for gran_time, gran_files in pair_file_names(fnames).items():
        gran_xobj = xarray.Dataset()
        for fname in gran_files:
            LOG.info(f"Opening: {fname}")
            # First grab metadata
            with xarray.open_dataset(fname) as ds:
                available_vars = list(ds.data_vars.keys())
                meta = xarray.Dataset(attrs=ds.attrs)
                try:
                    if "." in meta.attrs["time_coverage_start"]:
                        tformat = "%Y-%m-%dT%H:%M:%S.%fZ"
                    else:
                        tformat = "%Y-%m-%dT%H:%M:%SZ"
                    start_datetime = datetime.strptime(
                        meta.attrs["time_coverage_start"], tformat
                    )
                    end_datetime = datetime.strptime(
                        meta.attrs["time_coverage_end"], tformat
                    )
                except KeyError:
                    start_datetime = datetime.strptime(
                        meta.attrs["PGE_StartTime"], "%Y-%m-%d %H:%M:%S.%f"
                    )
                    end_datetime = datetime.strptime(
                        meta.attrs["PGE_EndTime"], "%Y-%m-%d %H:%M:%S.%f"
                    )
                if hasattr(meta, "SatelliteInstrument"):
                    pform = meta.attrs["SatelliteInstrument"]
                elif hasattr(meta, "satellite_name"):
                    pform = meta.attrs["satellite_name"]
                else:
                    raise AttributeError("Cannot identify platform name")
                geoips_attrs = {
                    "area_definition": area_def,
                    "start_datetime": start_datetime,
                    "end_datetime": end_datetime,
                    "vertical_data_type": "surface",
                    "source_name": "viirs",
                    "platform_name": get_geoips_platform_name(pform),
                    "data_provider": "nasa",
                    "interpolation_radius_of_influence": 2000,
                }
                meta.attrs = dict(meta.attrs, **geoips_attrs)

            if metadata_only is True:
                LOG.info(
                    "metadata_only is True, reading only first"
                    + "file for metadata information and returning"
                )
                return {"METADATA": meta}

            group = None
            if "Geolocation" in meta.title:
                # This is a geolocation file, need to open the group to
                # find the available datasets
                group = "geolocation_data"
                with xarray.open_dataset(fname, group=group) as ds:
                    available_vars = list(ds.data_vars.keys())

            if not chans:
                # Load all datasets if no chans are requested
                default_chans = available_vars
            else:
                default_chans = chans

            # Only load the requested variables
            drop_vars = list(set(available_vars).difference(set(default_chans)))

            # Use xarray to open the file
            xr = xarray.load_dataset(fname, group=group, drop_variables=drop_vars)

            empty_vars = []

            for xr_dim_name, xr_dim_val in xr.dims.items():
                for gr_dim_name, gr_dim_val in gran_xobj.dims.items():
                    if (xr_dim_name != gr_dim_name) & (xr_dim_val == gr_dim_val):
                        LOG.info(
                            "Detected %s shares the same dim size as %s",
                            xr_dim_name,
                            gr_dim_name,
                        )
                        LOG.info(
                            "Renaming to share same dimension name: %s", gr_dim_name
                        )
                        xr = xr.rename({xr_dim_name: gr_dim_name})

            for var, data in xr.variables.items():
                if data.size == 0:
                    LOG.info("Empty variable: %s. Dropping", var)
                    empty_vars.append(var)
                    continue
                # Check if time array exists
                time_key = "datetime_" + "_".join(data.dims)
                if (time_key not in gran_xobj) & (time_key not in xr):
                    LOG.info("Creating time key: %s", time_key)
                    time_array = np.empty(data.shape).astype(datetime)
                    time_array[:] = start_datetime
                    xr[time_key] = (data.dims, time_array)
            xr = xr.drop_vars(empty_vars)

            gran_xobj = merge_xarrays(xr, gran_xobj)

        if "Latitude" in list(gran_xobj.variables.keys()):
            gran_xobj = gran_xobj.rename({"Latitude": "latitude"})
            gran_xobj = gran_xobj.rename({"Longitude": "longitude"})

        loaded_vars = list(gran_xobj.variables.keys())
        if not all(x in loaded_vars for x in chans):
            # sorted(loaded_vars) != sorted(chans):
            LOG.info(
                "Not all variables available for %s. "
                + "Not appending to final xarray Dataset",
                gran_time,
            )
            continue
        # Merge xarrays together
        xobj = merge_xarrays(gran_xobj, xobj)

    for key in xobj.data_vars.keys():
        xobj[key].data = xobj[key].to_masked_array()
    xobj.attrs = meta.attrs

    vars_2d = []
    vars_1d = []
    for key in xobj.data_vars.keys():
        dims = list(xobj[key].dims)
        if "time" in dims:
            dims.remove("time")
        if len(dims) == 1:
            vars_1d.append(key)
        elif len(dims) == 2:
            vars_2d.append(key)

    xobj_1d = xobj.drop_vars(vars_2d)
    xobj_2d = xobj.drop_vars(vars_1d)

    rename_vars = {}
    for geo_var in ["longitude", "latitude"]:
        if geo_var not in xobj_1d.data_vars.keys():
            for key in xobj_1d.data_vars.keys():
                if geo_var in key:
                    print(f"Renaming {key} to latitude")
                    rename_vars[key] = geo_var
                    xobj_1d[geo_var] = xobj_1d[key]

    return {"viirs_data_2d": xobj_2d, "viirs_data_1d": xobj_1d, "METADATA": meta}



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Product Defaults interface module."""

from geoips.interfaces.base import BaseYamlInterface


class ProductDefaultsInterface(BaseYamlInterface):
    """Default values that can be applied to products."""

    name = "product_defaults"


product_defaults = ProductDefaultsInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit tests for geoips/commandline/log_setup.py.

Tests the following functions:
- log_with_emphasis
- add_log_level (by importing geoips.logging)
"""

import pytest

import random
import string
from geoips import logging
from geoips.commandline.log_setup import log_with_emphasis

LOG = logging.getLogger(__name__)


def generate_random_string(length):
    """Generate a random string of length :param length."""
    return "".join(random.choices(string.ascii_letters, k=length))


def insert_word_like_spaces_to_string(string):
    """Modify the input string by inserting spaces to make "words" of length 2-8.

    Parameters
    ----------
    str (str):
        The input string to be modified.

    Returns
    -------
    str (str):
        The modified string with spaces inserted at random locations.

    Example
    -------
    >>> insert_word_like_spaces_to_string("HelloWorld")
    'He lloW or ld'
    """
    loc = random.randint(2, 3)
    while loc < len(string):
        string = string[:loc] + " " + string[loc + 1 :]
        loc += random.randint(2, 8)
    return string


def insert_random_string_randomly(s, length):
    """Insert a randomly generated string of a given length at a random position.

    Parameters
    ----------
    s : str
        The original string into which the random string will be inserted.
    length : int
        The length of the random string to be generated and inserted.

    Returns
    -------
    str
        The modified string with a randomly generated string of length `length`
        inserted at a random position.

    Example
    --------
    >>> insert_random_string_randomly("hello", 3)
    'helXyZlo'
    """
    position = random.randint(0, len(s) - 1)
    return s[position:] + generate_random_string(length) + s[position:]


def generate_random_messages(add_long_word=False):
    """Generate a random amount of messages with random length."""
    num_messages = 20
    messages = [
        insert_word_like_spaces_to_string(
            generate_random_string(random.randint(5, 110))
        )
        for _ in range(num_messages)
    ]
    if add_long_word:

        def f(s):
            return insert_random_string_randomly(s, 88)  # insert string 88 chars long

        return list(map(f, messages))
    else:
        return messages


@pytest.mark.parametrize("message", generate_random_messages())
def test_log_with_emphasis(message, caplog, test_all_lines_same_length=True):
    """Pytest function for testing the output of 'log_with_emphasis'.

    The expected output of log_with_emphasis looks similar to this:
    ***************
    ** hello     **
    ** howdy     **
    ** what's up **
    ***************
    """  # noqa: RST212
    # ignoring check because flake8 flags the codeblocks as underlines
    caplog.set_level(logging.INFO)
    log_with_emphasis(LOG.info, message)

    assert (  # top of box is formmated correctly
        "*" * 9
        in caplog.messages[0]  # three for borders, and min of 5 for string length
    )
    assert message[0:5] in caplog.text  # test for first section of text
    assert "\n" in caplog.text  # assert log output is multi-lined

    assert caplog.messages[1].startswith("** ")  # test for left side of box
    assert caplog.messages[1].endswith(" **")  # test for left side of box

    if test_all_lines_same_length:
        assert (
            len(set(map(len, caplog.messages[:-1]))) == 1  # last line is blank
        )  # all logged lines are the same length

    with pytest.raises(ValueError):
        log_with_emphasis(LOG.info, "")


@pytest.mark.parametrize("message", generate_random_messages(add_long_word=True))
def test_log_with_emphasis_long_word(message, caplog):
    """Pytest function for testing the output of 'log_with_emphasis'.

    The expected output of log_with_emphasis usually looks like this:
    ***************
    ** hello     **
    ** howdy     **
    ** what's up **
    ***************
    However, by default we don't wrap long words. If a "long word" (string of chars
    surrounded by spaces of length >74) is present, the expected output of
    log_with_emphasis looks like this:
    ***********************************************************************************
    ** hello                                                                         **
    ** howdy                                                                         **
    ** what's up                                                                     **
    ** this is a very long string that we are not going to match at the top or bottom because it really is too long ok I think this is long enough **
    ***********************************************************************************
    """  # noqa: E501,RST212
    # ignoring line length check, and the section underline check because
    # flake8 flags the codeblocks as underlines
    caplog.set_level(logging.INFO)
    log_with_emphasis(LOG.info, message)
    log_lines = caplog.messages[1 : len(caplog.text) - 1]

    # all logged lines are NOT the same length (because we didn't wrap)
    assert not (len(set(map(len, log_lines[:-1]))) == 1)

    # find at least one line that is longer than 80 chars (aka not wrapped)
    assert any(map(lambda line: len(line) > 80, log_lines))


def test_log_interactive_geoips(caplog):
    """Test log.interactive using logging from geoips."""
    log = logging.getLogger(__name__)

    log.interactive("FROM PYTEST")
    assert "FROM PYTEST" in caplog.text
    assert "INTERACTIVE" in caplog.text


# Commenting this out, unless we can identify a test that does not require having
# test datasets cloned. Should not require test datasets to test interactive
# log from a plugin.
# def test_log_interactive_from_directly_imported_plugin(caplog):
#     """Use ABI reader, which calls log.interactive.
#
#     It used to be that plugins loaded via get_plugin outside of GeoIPS could not call
#     log.interactive. They would raise an AttributeError. This is to ensure that
#     problem has been fixed and remains fixed.
#     """
#     from geoips.interfaces import readers
#
#     reader = readers.get_plugin("abi_netcdf")
#
#     test_data_dir = os.getenv("GEOIPS_TESTDATA_DIR")
#     if test_data_dir:
#         # Only run this test if GEOIPS_TESTDATA_DIR environment variable has been set
#         # on the machine that is running the test.
#         test_data_dir += "/test_data_noaa_aws/data/goes16/20200918/1950"
#         test_data_files = glob(test_data_dir + "/*C14*.nc")
#         _ = reader(test_data_files)
#         assert "INTERACTIVE" in caplog.text



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Code to implement GeoIPS Command Line Interface (CLI).

Will implement a plethora of commands, but for the meantime, we'll work on
'geoips list' and 'geoips run'
"""

from os.path import basename, dirname, join
import sys

from colorama import Fore, Style

from geoips.commandline.cmd_instructions import get_instructions
from geoips.commandline.geoips_command import GeoipsCommand
from geoips.commandline.geoips_config import GeoipsConfig
from geoips.commandline.geoips_describe import GeoipsDescribe
from geoips.commandline.geoips_list import GeoipsList
from geoips.commandline.geoips_run import GeoipsRun
from geoips.commandline.geoips_test import GeoipsTest
from geoips.commandline.geoips_tree import GeoipsTree
from geoips.commandline.geoips_validate import GeoipsValidate
from geoips.commandline.log_setup import setup_logging


class GeoipsCLI(GeoipsCommand):
    """Top-Level Class for the GeoIPS Commandline Interface (CLI).

    This class includes a list of Command Classes, which will implement the core
    functionality of the CLI. This includes the following as of right now:
    - [GeoipsConfig, GeoipsGet, GeoipsList, GeoipsRun, GeoipsTest, GeoipsValidate]
    """

    name = "geoips"  # Needed since we inherit from GeoipsCommand
    command_classes = [
        GeoipsConfig,
        GeoipsDescribe,
        GeoipsList,
        GeoipsRun,
        GeoipsTest,
        GeoipsTree,
        GeoipsValidate,
    ]

    def __init__(self, instructions_dir=None, legacy=False):
        """Initialize the GeoipsCLI and each of it's command classes.

        The CLI contains a single top-level argparse.ArgumentParser() which contains
        subparsers related to each subcommand. This ensures that each command has a
        unique set of arguments inheirted from command -> subcommand -> subcommand,
        and so on. For example, the GeoipsList Command Class's arguments are inherited
        by all subcommand child classes, which recursively can have their own child
        subcommand classes.

        Parameters
        ----------
        instructions_dir: str or Posix.Path
            - The path to the directory which includes the commandline instructions.
              This is only used for testing purposes so we can ensure the correct
              functionality occurs for possibly missing / invalid instruction files.
        """
        if instructions_dir:
            # Instructions dir has been provided, use the instructions found in that
            # directory so we can test that the correct functionality occurs for any
            # given instruction file state.
            self.cmd_instructions = get_instructions(instructions_dir)
        else:
            # Otherwise use the default instructions which we know are correct
            # (and if they're not, the appropriate error will be raised.)
            self.cmd_instructions = None

        # parse_known_args expects arguments in a specific order. So, currrently,
        # 'geoips --log-level info <rest of command>' will work but
        # 'geoips <rest of command> --log-level info' will not. The functionality below
        # rearranges the log level arguments to match the working version. This way,
        # users can add --log-level <log_level_name> anywhere in the command and it will
        # work.
        if set(sys.argv).intersection(set(["--log-level", "-log"])):
            # One of the flags was found in the arguments provided
            log_idx = max(
                [
                    idx if arg in ["--log-level", "-log"] else -1
                    for idx, arg in enumerate(sys.argv)
                ]
            )
            # Make sure that the argument list is long enough for log level to be
            # provided. It doesn't have to be correct, that validation will be done
            # by argparse
            if len(sys.argv) > log_idx + 1:
                flag = sys.argv[log_idx]
                log_level = sys.argv[log_idx + 1]
                # Get the flag and log_level, remove them from the argument list, and
                # insert them in working locations.
                # I.e. geoips --log-level <log_level_name>
                sys.argv.pop(log_idx + 1)
                sys.argv.pop(log_idx)
                sys.argv.insert(1, log_level)
                sys.argv.insert(1, flag)

        super().__init__(legacy=legacy)

    def execute_command(self):
        """Execute the given command."""
        self.GEOIPS_ARGS = self.parser.parse_args()
        if hasattr(self.GEOIPS_ARGS, "exe_command"):
            # The command called is executable (child of GeoipsExecutableCommand)
            # so execute that command now.
            self.GEOIPS_ARGS.exe_command(self.GEOIPS_ARGS)
            return self.GEOIPS_ARGS
        else:
            print(
                # f'"{self.GEOIPS_ARGS.command}" command requires a subcommand.\n\n'
                f"{self.GEOIPS_ARGS.command_parser.format_usage()}"
            )
            sys.exit(2)


def deprecate_create_plugin_registries():
    """Deprecate the 'create_plugin_registries' command if it was supplied.

    If 'create_plugin_registries' was called, convert that command to
    'geoips config create-registries' and print a deprecation warning. These two
    commands are equivalent, however we are transitioning to the sole use of the CLI
    rather than unconnected console scripts.
    """
    if basename(sys.argv[0]) == "create_plugin_registries":
        LOG = setup_logging(logging_level="warning")
        LOG.warning(
            msg=(
                "'create_plugin_registries' is deprecated. Please use 'geoips config "
                "create-registries' from now on. This functionality will be removed in "
                "the future."
            ),
            stacklevel=2,
        )
        sys.argv[0] = join(dirname(sys.argv[0]), "geoips")
        sys.argv.append("config")
        sys.argv.append("create-registries")


def support_legacy_procflows():
    """Run a series of checks on sys.argv to support legacy calls to run_procflow.

    The new GeoIPS CLI has changed how we run our process workflows from now on. Instead
    of calling 'run_procflow' or 'data_fusion_procflow', we now call
    'geoips run single_source' or 'geoips run data_fusion'. This function parses through
    sys.argv and performs the necessary translations to match our current format so that
    the CLI's argparser can call the appropriate functionality.

    Returns
    -------
    legacy: bool
        - The truth value as to whether or not a legacy procflow call was used
    """
    defined_procflow = None
    # Including '-h' here as we need to be able to support help messages for this cmd
    supported_procflows = ["config_based", "data_fusion", "single_source", "-h"]
    if (
        basename(sys.argv[0]) == "geoips"
        and len(sys.argv) > 2
        and sys.argv[1] == "run"
        and (len(sys.argv) < 3 or sys.argv[2] not in supported_procflows)
    ):
        # Either a procflow was not specified or it was an invalid procflow. Notify the
        # user of that with a 'NotImplementedError'.
        if len(sys.argv) >= 3:
            procflow_name = sys.argv[2]
        else:
            procflow_name = "No procflow supplied."
        raise NotImplementedError(
            f"'geoips run' was called alongside procflow: '{procflow_name}'.\nIf you "
            "did not supply a procflow name, this is not supported currently.\n"
            "Eventually, 'geoips run' will call the 'order_based' procflow, however "
            "this is not at the current time of use.\nFor a list of supported "
            f"procflows, choose one of the following: {supported_procflows}."
        )
    elif basename(sys.argv[0]) == "run_procflow":
        entrypoint = "run_procflow"
        defined_procflow = "single_source"

    elif basename(sys.argv[0]) == "data_fusion_procflow":
        entrypoint = "data_fusion_procflow"
        defined_procflow = "data_fusion"

    if defined_procflow:
        # Either 'run_procflow' or 'data_fusion_procflow' was called, make the
        # appropriate translations in sys.argv so the CLI's parser can call the correct
        # functionality
        if "--procflow" in sys.argv:
            # If '--procflow' was found in the command line arguments, loop through the
            # arguments and grab the correct procflow.
            for idx, arg in enumerate(sys.argv):
                if arg == "--procflow" and len(sys.argv) > idx + 1:
                    defined_procflow = sys.argv[idx + 1]
        sys.argv[0] = sys.argv[0].replace(entrypoint, "geoips")
        sys.argv.insert(1, "run")
        sys.argv.insert(2, defined_procflow)
        return True
    return False


def print_beta_warning():
    """Notify the user that the CLI is still in Beta development stage."""
    print(
        Fore.RED
        + "\nWARNING: "
        + Fore.YELLOW
        + "The GeoIPS CLI is currently under development and is subject "
        "to change.\nUntil this warning is removed, do not rely on the CLI to be "
        "static.\nPlease feel free to test the CLI and report any bugs or comments as "
        "an issue here:\n"
        + Fore.BLUE
        + "https://github.com/NRLMMD-GEOIPS/geoips/issues/new/choose\n"
        + Style.RESET_ALL
    )


def main(suppress_args=True):
    """Entry point for GeoIPS command line interface (CLI).

    Parameters
    ----------
    suppress_args: bool, (default=True)
        - Whether or not we want to suppress the arguments returned from the CLI.
          If False, return the argument namespace. Used for unit testing.
    """
    # Support legacy 'run_procflow' calls if applicable
    legacy = support_legacy_procflows()
    # Deprecate 'create_plugin_registries' if called
    deprecate_create_plugin_registries()
    # Initialize the CLI and all of its commands
    geoips_cli = GeoipsCLI(legacy=legacy)
    # Execute the called command
    args = geoips_cli.execute_command()
    # Notify that the user is in Beta development status right now.
    print_beta_warning()
    if not suppress_args:
        return args


if __name__ == "__main__":
    main()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Clean TC filename production (no backgrounds or overlays)."""

from geoips.filenames.base_paths import PATHS as gpaths

interface = "filename_formatters"
family = "standard"
name = "tc_clean_fname"


def call(
    area_def,
    xarray_obj,
    product_name,
    coverage=None,
    output_type="png",
    output_type_dir=None,
    product_dir=None,
    product_subdir=None,
    source_dir=None,
    basedir=gpaths["TCWWW"],
    output_dict=None,
):
    """Clean TC product filename formatter (no gridlines, titles, etc).

    This ensures output ends up in "png_clean" directory, with "-clean"
    appended to the extra field, to avoid conflict with tc_fname based
    annotated imagery.  Uses "tc_fname" module as a base.

    Parameters
    ----------
    area_def : pyresample AreaDefinition
        Contains metadata regarding sector
    xarray_obj : xarray Dataset
        Contains metadata regarding dataset
    product_name : str
        String product_name specification for use in filename
    coverage : float
        Percent coverage, for use in filename
    output_type : str, optional
        Requested output format, ie png, jpg, tif, etc, defaults to None.
    output_type_dir : str, optional
        Directory name for given output type (ie png_clean, png, etc), defaults
        to None.
    product_dir : str, optional
        Directory name for given product, defaults to None.
    product_subdir : str, optional
        Subdir name for given product, if any, defaults to None.
    source_dir : str, optional
        Directory name for given source, defaults to None.
    basedir : str, optional
        Base directory, defaults to $TCWWW.

    Returns
    -------
    str
        Full path to output "clean" filename - with "-clean" appended to
        extra field, and "_clean" appended to output_type_dir.
    """
    from geoips.interfaces import filename_formatters

    return filename_formatters.get_plugin("tc_fname")(
        area_def,
        xarray_obj,
        product_name,
        coverage,
        output_type=output_type,
        output_type_dir=output_type + "_clean",
        product_dir=product_dir,
        product_subdir=product_subdir,
        source_dir=source_dir,
        basedir=basedir,
        extra_field="clean",
        output_dict=output_dict,
    )



"""Pydantic models used to validate GeoIPS sector plugins."""

from datetime import datetime
from enum import Enum
from typing import Literal, Tuple, Union
from typing_extensions import Annotated

from pydantic import (
    Field,
    ConfigDict,
    field_validator,
)
from pydantic.functional_validators import AfterValidator

from geoips.pydantic.bases import FrozenModel, PermissiveFrozenModel, PluginModel


class EarthConstants(float, Enum):
    """A simple class containing constant values relative to the geometry of Earth."""

    SEMI_MAJOR_AXIS = 6371228.0  # Semimajor axis of the earth in meters.


class XYCoordinate(FrozenModel):
    """A coordinate in projection units."""

    x: float = Field(None, description="The x coordinate in projection units.")
    y: float = Field(None, description="The y coordinate in projection units.")


def lat_lon_coordinate(arg: tuple[float, float]) -> tuple[float, float]:
    """Validate a latitude and longitude coordinate."""
    if arg[0] < -90 or arg[0] > 90:
        raise ValueError("Latitude must be between -90 and 90")
    if arg[1] < -180 or arg[1] > 180:
        raise ValueError("Longitude must be between -180 and 180")
    return arg


LatLonCoordinate = Annotated[Tuple[float, float], AfterValidator(lat_lon_coordinate)]


class SectorProjection(PermissiveFrozenModel):
    """Projection information for a sector.

    This is a dictionary that provides Proj projection information for the sector. For
    more information on what parameters can be supplied, see the Proj documentation.

    Validation has only been implemented for some of the most common options. If you
    need validation for a parameter that is not currently implemented, please open an
    issue and, if possible, a pull request on GitHub.
    """

    proj: str = Field(..., description="Proj projection alias.")
    a: EarthConstants = Field(
        EarthConstants.SEMI_MAJOR_AXIS,
        description="Semimajor axis of the ellipsoid in meters.",
    )
    R: Annotated[
        float,
        Field(
            None,
            strict=True,
            ge=0,
            description=(
                "Radius of the sphere, given in meters. If used in conjunction with "
                "``+ellps``, :option:`+R` takes precedence."
                "See https://proj.org/en/stable/usage/ellipsoids.html#ellipsoid-size-parameters "  # NOQA
                "for more information."
            ),
        ),
    ]
    ellps: str = Field(
        "GRS80",
        description=(
            "The name of a built-in ellipsoid definition. "
            "See https://proj.org/en/stable/usage/ellipsoids.html#built-in-ellipsoid-definitions "  # NOQA
            " for more information, or execute :option:`proj -le` for a list of "
            "built-in ellipsoid names."
            "*Defaults to 'GRS80'.*"
        ),
    )
    h: Annotated[
        float,
        Field(
            None,
            strict=True,
            ge=0,
            description=(
                "Height of the view point above the Earth and must be in the same units"
                " as the radius of the sphere or semimajor axis of the ellipsoid."
            ),
        ),
    ]
    k_0: Annotated[
        float,
        Field(
            1.0,
            strict=True,
            ge=0,
            description=(
                "Scale factor. Determines scale factor used in the projection."
                "*Defaults to 1.0.*"
            ),
        ),
    ]
    lat_0: Annotated[
        float,
        Field(
            None,
            strict=True,
            ge=-90,
            le=90,
            description="Latitude of origin in degrees.",
        ),
    ]
    lat_1: Annotated[
        float,
        Field(
            0.0,
            strict=True,
            ge=-90,
            le=90,
            description=("First standard parallel." "*Defaults to 0.0.*"),
        ),
    ]
    lat_2: Annotated[
        float,
        Field(
            0.0,
            strict=True,
            ge=-90,
            le=90,
            description=("Second standard parallel." "*Defaults to 0.0.*"),
        ),
    ]
    lat_ts: Annotated[
        float,
        Field(
            0.0,
            strict=True,
            ge=0,
            description=(
                "Latitude of true scale. Defines the latitude where scale is not "
                "distorted."
                "Takes precedence over ``+k_0`` if both options are used together."
                "*Defaults to 0.0.*"
            ),
        ),
    ]
    lon_0: Annotated[
        float,
        Field(
            None,
            strict=True,
            ge=-180,
            le=180,
            description="Longitude of origin in degrees.",
        ),
    ]
    t_epoch: float = Field(None, description=("Central epoch of the transformation."))
    t_final: float = Field(
        None,
        description=(
            "Final epoch that the coordinate will be propagated to after transformation."  # NOQA
            "The special epoch *now* can be used instead of writing a specific period in"  # NOQA
            "time. When *now* is used, it is replaced internally with the epoch of the "
            "transformation. This means that the resulting coordinate will be slightly "
            "different if carried out again at a later date."
        ),
    )
    x_0: float = Field(
        0.0,
        description=(
            "False easting, easting at false origin or easting at projection centre "
            "(naming and meaning depend on the projection method). Always in meters."
            "*Defaults to 0.0.*"
        ),
    )
    y_0: float = Field(
        0.0,
        description=(
            "False northing, northing at false origin or northing at projection centre "
            "(naming and meaning depend on the projection method). Always in meters."
            "*Defaults to 0.0.*"
        ),
    )
    units: Literal["m", "km", "degrees"] = Field(
        "m",
        description=(
            "Units of the projection. "
            "This should not need to be changed from the default. "
            "This controls the units of the x and y coordinates in the projection but "
            "has no impact on the units of the resolution or area_extent because we "
            "specify those units separately."
        ),
    )


# Ex: Change inheritance to FrozenModel
class SectorShape(FrozenModel):
    """The shape of the sector in pixels."""

    height: Annotated[
        int,
        Field(
            ...,
            strict=True,
            gt=0,
            description="The height of the sector in pixels. Must be greater than 0.",
        ),
    ]
    width: Annotated[
        int,
        Field(
            ...,
            strict=True,
            gt=0,
            description="The width of the sector in pixels. Must be greater than 0.",
        ),
    ]


class SectorResolution(FrozenModel):
    """The resolution of the sector in projection units.

    The height and width of pixels in the units specified by the sector's projection
    units.
    """

    dx: Annotated[
        float,
        Field(
            ...,
            strict=True,
            gt=0,
            description=(
                "The width of pixels in the units specified by "
                "the sector's projection units."
            ),
        ),
    ]
    dy: Annotated[
        float,
        Field(
            ...,
            strict=True,
            gt=0,
            description=(
                "The height of pixels in the units specified by "
                "the sector's projection units."
            ),
        ),
    ]


class SectorAreaExtent(FrozenModel):
    """The extent of the sector in projection units.

    For more information on how this is used, see the pyresample documentation.
    """

    lower_left_xy: Tuple[int, int] = Field(
        ...,
        description="Lower left corner of the sector in projection units.",
    )
    upper_right_xy: Tuple[int, int] = Field(
        ...,
        description="Upper right corner of the sector in projection units. ",
    )


# This possibly could be PermissiveFrozenModel
class AreaDefinitionSpec(FrozenModel):
    """Defines an AreaDefinition for use with pyresample.

    The resulting dictionary should be able to just be passed to
    ``pyresample.create_area_def()``.
    """

    area_id: str = Field(
        None,
        description=(
            "A name for the resulting pyresample AreaDefinition. "
            "Defaults to the sector's name."
        ),
    )
    description: str = Field(
        None,
        description=(
            "A description for the resulting pyresample AreaDefinition. "
            "Defaults to the sector's docstring."
        ),
    )
    projection: SectorProjection = Field(
        ...,
        description=(
            "A dictionary providing Proj projection information for the sector. "
            "For more information please see the Proj documentation."
        ),
    )
    # This can either be a tuple of integers (preferred) or a
    # SectorShape object (deprecated)
    shape: Union[
        Tuple[
            Annotated[int, Field(strict=True, gt=0)],
            Annotated[int, Field(strict=True, gt=0)],
        ],
        SectorShape,
    ] = None
    area_extent: Union[Tuple[float, float, float, float], SectorAreaExtent] = Field(
        None,
        description=(
            "Sector area extent in projection units. "
            "For more information see the pyresample documentation."
        ),
    )
    resolution: Union[
        Annotated[float, Field(strict=True, gt=0)],
        Tuple[
            Annotated[float, Field(strict=True, gt=0)],
            Annotated[float, Field(strict=True, gt=0)],
        ],
        SectorResolution,
    ] = Field(
        None,
        description=(
            "The size of the pixels in the sector in projection units. "
            "May be specified as a single float or a tuple of two floats "
            "describing the resolution in the x and y directions separately."
            "See the pyresample documentation for more information."
        ),
    )
    units: Literal["m", "km", "meters", "kilometers", "deg", "degrees"] = Field(
        "m",
        description=(
            "The units used for resolution and area_extent. "
            "This takes priority over the units specified in the projection. "
            "For more information on this parameter and its priority order, see the "
            "pyresmaple documentation."
        ),
    )

    center: Union[Tuple[float, float], XYCoordinate] = Field(
        (0, 0),
        description=(
            "The center of the sector in projection units. Defaults to (0, 0). "
            "See the pyresample documentation for more information."
        ),
    )


class RegionMetadata(FrozenModel):
    """Metadata format for standard static sectors."""

    model_config = ConfigDict(coerce_numbers_to_str=False)

    continent: str = Field(..., description="Continent which the sector resides in.")
    country: str = Field(..., description="Country which the sector resides in.")
    area: str = Field(..., description="Geographic area of the sector.")
    subarea: str = Field(..., description="Geographic subarea of the sector.")
    state: str = Field(..., description="State which the sector resides in.")
    city: str = Field(..., description="City which the sector resides in.")


class StaticMetadata(FrozenModel):
    """Metadata format for standard static sectors.

    This is the same as StaticMetadata, just with an additional 'region' level. This is
    a convenience model for specifying static sector plugins in a legacy format.
    """

    region: RegionMetadata = Field(
        ..., description="Additional field used to specify metadata in a legacy format."
    )


class BoxMetadata(FrozenModel):
    """Metadata format for pyroCb sectors."""

    min_lat: Annotated[
        float,
        Field(
            ...,
            strict=True,
            ge=-90,
            le=90,
            description="Bottom left latitude in degrees.",
        ),
    ]
    min_lon: Annotated[
        float,
        Field(
            ...,
            strict=True,
            ge=-180,
            le=180,
            description="Bottom left longitude in degrees.",
        ),
    ]
    max_lat: Annotated[
        float,
        Field(
            ...,
            strict=True,
            ge=-90,
            le=90,
            description="Upper right latitude in degrees.",
        ),
    ]
    max_lon: Annotated[
        float,
        Field(
            ...,
            strict=True,
            ge=-180,
            le=180,
            description="Upper right longitude in degrees.",
        ),
    ]

    box_resolution_km: Annotated[
        float,
        Field(
            ...,
            strict=True,
            gt=0,
            description="The resolution of each pixel in kilometers.",
        ),
    ]


class StitchedMetadata(StaticMetadata):
    """Metadata for stitched imagery sectors."""

    primary_area_definition: str = Field(
        ..., description="Name of the area definition to be used."
    )


class TCMetadata(FrozenModel):
    """Metdata format for Tropical Cyclone sectors."""

    pressure: Annotated[
        float,
        Field(
            ..., strict=True, gt=0, description="Pressure of the storm in millibars."
        ),
    ]
    vmax: Annotated[
        float,
        Field(
            ...,
            strict=True,
            gt=0,
            description="Maximum velocity in knots of the storm.",
        ),
    ]
    clat: Annotated[
        float,
        Field(
            ..., strict=True, ge=-90, le=90, description="Center latitude of the storm."
        ),
    ]
    clon: Annotated[
        float,
        Field(
            ...,
            strict=True,
            ge=-180,
            le=180,
            description="Center longitude of the storm.",
        ),
    ]
    synoptic_time: datetime = Field(..., description="Synoptic time of the storm.")
    aid_type: str = Field(..., description="Tropical cyclone forecast aid category.")
    storm_num: Annotated[
        int,
        Field(
            ...,
            strict=True,
            gt=0,
            description=(
                "Two digit storm number in sequential order from the start of the "
                "hurricane season."
            ),
        ),
    ]
    storm_name: str = Field(description="Name of the storm.")
    storm_basin: str = Field(
        ...,
        description=(
            "Two character representation of the basin in which the storm originated "
            "in."
        ),
    )
    storm_year: Annotated[
        int,
        Field(
            ...,
            strict=True,
            gt=1900,
            description="Year in which the storm originated, after 1900.",
        ),
    ]
    deck_line: str = Field(
        ...,
        description=(
            "Deck data file used within the weather forecasting system for this storm."
        ),
    )
    source_file: str = Field(..., description="File in which the storm data came from.")
    final_storm_name: str = Field(
        ...,
        description="Final name of the storm used within the track file.",
    )


class VolcanoMetadata(FrozenModel):
    """Metadata format for Volcano sectors."""

    summit_elevation: float = Field(
        ..., description="Elevation of the volcano's summit."
    )
    plume_height: float = Field(..., description="Altitude of the volcanic plume.")
    wind_speed: Annotated[
        float,
        Field(..., strict=True, gt=0, description="Windspeed of the volcano sector."),
    ]
    wind_dir: Annotated[
        float,
        Field(
            ...,
            strict=True,
            ge=0,
            le=360,
            description="Angular direction of wind within the volcano sector.",
        ),
    ]
    clat: Annotated[
        float,
        Field(
            ...,
            strict=True,
            ge=-90,
            le=90,
            description="Center latitude of the volcano sector.",
        ),
    ]
    clon: Annotated[
        float,
        Field(
            ...,
            strict=True,
            ge=-180,
            le=180,
            description="Center longitude of the volcano sector.",
        ),
    ]


class SectorPluginModel(PluginModel):
    """Sector plugin format."""

    model_config = ConfigDict(extra="allow")
    spec: AreaDefinitionSpec = Field(
        ...,
        description=(
            "Specification of the sector's geographical domain. "
            "Used to generate a pyresample AreaDefinition."
        ),
    )
    metadata: Union[
        BoxMetadata,
        StaticMetadata,
        StitchedMetadata,
        TCMetadata,
        VolcanoMetadata,
    ] = Field(
        ...,
        description=(
            "Metadata describing the sector "
            "(mostly used in FilenameFormatters and MetadataFormatters)."
        ),
    )

    @field_validator("metadata", mode="before")
    @classmethod
    def coerce_metadata(cls, value):
        """Coerce metadata to StaticMetadata if possible.

        Doing this as pydantic by default will return the last class that validates
        correctly, and this applies to other metadata classes.
        """
        if isinstance(value, dict) and "region" in value:
            return StaticMetadata(**value)
        return value

    @field_validator("spec", mode="before")
    @classmethod
    def add_parent(cls, value, info):
        """Collect default area_id and description from parent.

        This is a field validator that runs before the root validator. It accesses the
        properties of the parent object to fill in area_id and description if they are
        not provided. These values come from the parent's name and docstring.
        """
        if "area_id" not in value or not value["area_id"]:
            value["area_id"] = info.data["name"]
        if "description" not in value or not value["description"]:
            value["description"] = info.data["docstring"]
        return value
